import xtrack as xt
import numpy as np

env = xt.load_madx_lattice('../../test_data/fcc_ee/fccee_z.seq')
line = env.fccee_p_ring
line.particle_ref = xt.Particles(p0c=45.6e9, mass0=xt.ELECTRON_MASS_EV)

tt = line.get_table()

# I want to add a random octupolar error to all quadrupoles whose name starts
# with 'qf'

# Table with all quads
tt_quad = tt.rows[tt.element_type=='Quadrupole']
tt_selected = tt_quad.rows['qf.*']
# is:
# Table: 752 rows, 11 cols
# name                s element_type isthick isreplica ...
# qfg2.1        468.213 Quadrupole      True     False
# qfg2.2        520.504 Quadrupole      True     False
# qfg2.3        572.795 Quadrupole      True     False
# qfg2.4        625.085 Quadrupole      True     False
# qfg2.5        677.376 Quadrupole      True     False
# qfg2.6        729.667 Quadrupole      True     False
# qfg2.7        781.957 Quadrupole      True     False
# qfg2.8        834.248 Quadrupole      True     False
# qfg2.9        886.539 Quadrupole      True     False
# qfg2.10        938.83 Quadrupole      True     False
# ...
# qf2.351         88811 Quadrupole      True     False
# qf2.352       88915.2 Quadrupole      True     False
# qf2.353       89019.4 Quadrupole      True     False

# Generate random octupolar errors
k3l_rel_rms = 1e-3
k3l_rel = k3l_rel_rms * (np.random.rand(len(tt_selected)))

# Generate error dictionary
errors = {}
for ii, nn in enumerate(tt_selected.name):
    errors[nn] = {'rel_knl': [0, 0, k3l_rel[ii]]}

# Set errors in the line
env.set_multipolar_errors(errors)

# The number of multipolar kicks along the elements is set automatically
# (one kick upstream by default for quadrupoles). In can be changed as follows
env.set(tt_selected.name, num_multipole_kicks=4)

# Inspect one element:
env['qfg2.5'].knl
# is : [0.00000000e+00, 0.00000000e+00, 4.02838126e-05, 0.00000000e+00,
#       0.00000000e+00, 0.00000000e+00]
env['qfg2.5'].num_multipole_kicks
# is: 4



import numpy as np
import pytest

import xtrack as xt
import xpart as xp


def test_seeds():
    # Define a reference particle
    particle_ref = xt.Particles(mass0=xt.ELECTRON_MASS_EV, q0=1, p0c=4e9)

    # Create a particle object containing some particles with capacity twice the number of particles
    n_part = 10
    particles = xp.build_particles(particle_ref=particle_ref, num_particles=n_part, _capacity=2*n_part)

    # Explicitly initialize the random number generator for the particle object
    seeds = np.random.randint(0, np.iinfo(np.uint32).max, particles._capacity)
    particles._init_random_number_generator(seeds=seeds)

    # Save the seeds for the random number generator
    rng_s1_before = particles._rng_s1.copy()
    rng_s2_before = particles._rng_s2.copy()
    rng_s3_before = particles._rng_s3.copy()
    rng_s4_before = particles._rng_s4.copy()

    # Manually kill some particles
    n_part_to_kill = np.random.randint(1, n_part+1)
    idx_part_to_kill = np.random.choice(n_part, n_part_to_kill, replace=False)

    particles.state[idx_part_to_kill] = -1
    particles._num_active_particles = sum(particles.state == 1)
    particles._num_lost_particles = n_part - particles._num_lost_particles

    # Create some new particles
    n_part_new = np.random.randint(1, particles._capacity - n_part)
    new_particles = xp.build_particles(particle_ref=particle_ref, num_particles=n_part_new)

    # Add the new particles to the main particle object
    particles.add_particles(new_particles)

    # Save the seeds for the random number generator after the new particles addition
    rng_s1_after = particles._rng_s1
    rng_s2_after = particles._rng_s2
    rng_s3_after = particles._rng_s3
    rng_s4_after = particles._rng_s4

    # Verify that the seeds for the random number generator stay the same
    assert np.array_equal(rng_s1_after, rng_s1_before)
    assert np.array_equal(rng_s2_after, rng_s2_before)
    assert np.array_equal(rng_s3_after, rng_s3_before)
    assert np.array_equal(rng_s4_after, rng_s4_before)


from xtrack.pipeline.core import PipelineID, PipelineCommunicator
from xtrack.general import _print

class PipelineManager:
    def __init__(self,communicator = None):
        self._IDs = {}
        self._particles_per_rank = {}
        self._elements = {}
        self._pending_requests = {}
        self._last_request_turn = {}

        self.verbose = False

        if communicator is not None:
            self._communicator = communicator
        else:
            self._communicator = PipelineCommunicator()

    def add_particles(self,particles_name,rank):
        if rank in self._particles_per_rank.keys():
            pipeline_number = len(self._particles_per_rank[rank])
        else:
            pipeline_number = 0
            self._particles_per_rank[rank] = []
        pipeline_ID = PipelineID(rank,pipeline_number)
        self._IDs[particles_name] = pipeline_ID
        self._particles_per_rank[rank].append(particles_name)

    def get_particles_ID(self,particles_name):
        return self._IDs[particles_name]

    def get_particles_rank(self,particles_name):
        return self._IDs[particles_name].rank

    def add_element(self,element_name):
        self._elements[element_name] = len(self._elements)

    #
    # The tag is a int that identifies messages given the rank of the sender and of the receiver
    #
    def get_message_tag(self,element_name,sender_name,receiver_name,internal_tag=0):
        tag = self._elements[element_name] + len(self._elements)*self._IDs[sender_name].number + len(self._elements)*len(self._IDs)*self._IDs[receiver_name].number + len(self._elements)*len(self._IDs)*len(self._IDs)*internal_tag
        #if tag > self._max_tag:
        #    _print(f'PyPLINEDElement WARNING {self.name}: MPI message tag {tag} is larger than max ({self._max_tag})')
        return tag

    #
    # The key is string that uniquely identifies a message
    #
    def get_message_key(self,element_name,sender_name,receiver_name,tag=0):
        return f'{element_name}_{sender_name}_{receiver_name}_{tag}'

    def is_ready_to_send(self,element_name,sender_name,receiver_name,turn,internal_tag=0):
        tag = self.get_message_tag(element_name=element_name,sender_name=sender_name,receiver_name=receiver_name,internal_tag=internal_tag)
        key = self.get_message_key(element_name=element_name,sender_name=sender_name,receiver_name=receiver_name,tag=tag)
        if key not in self._last_request_turn.keys():
            return True
        if turn <= self._last_request_turn[key]:
            if self.verbose:
                _print(f'Pipeline manager {element_name}: {sender_name} at rank {self.get_particles_rank(sender_name)} has already sent a message to {receiver_name} at rank {self.get_particles_rank(receiver_name)} at turn {turn} with tag {tag}')
            return False
        if not self._pending_requests[key].Test():
            if self.verbose:
                _print(f'Pipeline manager {element_name}: {sender_name} at rank {self.get_particles_rank(sender_name)} previous message to {receiver_name} at rank {self.get_particles_rank(receiver_name)} with tag {tag} was not receviced yet')
            return False
        return True

    def send_message(self,send_buffer,element_name,sender_name,receiver_name,turn,internal_tag=0):
        tag = self.get_message_tag(element_name=element_name,sender_name=sender_name,receiver_name=receiver_name,internal_tag=internal_tag)
        key = self.get_message_key(element_name=element_name,sender_name=sender_name,receiver_name=receiver_name,tag=tag)
        self._last_request_turn[key] = turn
        if self.verbose:
            _print(f'Pipeline manager {element_name}: {sender_name} at rank {self.get_particles_rank(sender_name)} sending message to {receiver_name} at rank {self.get_particles_rank(receiver_name)} at turn {turn} with tag {tag}')
        self._pending_requests[key] = self._communicator.Issend(send_buffer,dest=self.get_particles_rank(receiver_name),tag=tag)

    def is_ready_to_receive(self,element_name,sender_name,receiver_name,internal_tag=0):
        tag = self.get_message_tag(element_name=element_name,sender_name=sender_name,receiver_name=receiver_name,internal_tag=internal_tag)
        is_ready = self._communicator.Iprobe(source=self.get_particles_rank(sender_name), tag=tag)
        if self.verbose and not is_ready:
            _print(f'Pipeline manager {element_name}: {receiver_name} at rank {self.get_particles_rank(receiver_name)} is not ready to receive from {sender_name} at rank {self.get_particles_rank(sender_name)} with tag {tag}')
        return is_ready

    def receive_message(self,receive_buffer,element_name,sender_name,receiver_name,internal_tag=0):
        tag = self.get_message_tag(element_name=element_name,sender_name=sender_name,receiver_name=receiver_name,internal_tag=internal_tag)
        if self.verbose:
            _print(f'Pipeline manager {element_name}: {receiver_name} at rank {self.get_particles_rank(receiver_name)} recieving from {sender_name} at rank {self.get_particles_rank(sender_name)} with tag {tag}')
        self._communicator.Recv(receive_buffer,source=self.get_particles_rank(sender_name),tag=tag)



import xtrack as xt
import numpy as np
from time import time

# rng = xt.RandomUniform()
rng = xt.RandomUniformAccurate()


n_gen = 0
min_val = 2.
max_val = -2.
min_non_zero = 2
n_zeros = 0
t0 = time()
while True:
    xx = np.squeeze(rng.generate(n_seeds=1, n_samples=1e8))

    min_val = min(min_val, xx.min())
    max_val = max(max_val, xx.max())
    min_non_zero = min(min_non_zero, xx[xx!=0].min())
    n_zeros += len(xx[xx==0])
    n_gen += len(xx)

    print(f'T={time()-t0:.2f}s, '
          f'Generated {n_gen:.4e} numbers, '
          f'max: 1 - {1 - max_val:.2e}, '
          f'min: {min_val:.2e}, '
          f'min non zero: {min_non_zero:.2e}, '
          f'n zeros: {n_zeros}')



# copyright ############################### #
# This file is part of the Xtrack Package.  #
# Copyright (c) CERN, 2021.                 #
# ######################################### #
import json
import pathlib

import numpy as np
import pytest
from cpymad.madx import Madx

import xobjects as xo
import xpart as xp
import xtrack as xt
from xobjects.test_helpers import for_all_test_contexts, fix_random_seed

test_data_folder = pathlib.Path(
    __file__).parent.joinpath('../test_data').absolute()


@for_all_test_contexts
def test_simple_collective_line(test_context):
    num_turns = 100
    elements = [xt.Drift(length=2., _context=test_context) for _ in range(5)]
    elements[3].iscollective = True
    line = xt.Line(elements=elements)
    line.reset_s_at_end_turn = False

    particles = xp.Particles(x=[1e-3, 2e-3, 3e-3], p0c=7e12,
            _context=test_context)
    line.build_tracker(_context=test_context)
    line.track(particles, num_turns=num_turns)

    particles.move(_context=xo.ContextCpu())

    assert np.all(particles.at_turn == num_turns)
    xo.assert_allclose(particles.s, 10 * num_turns, rtol=0, atol=1e-14)



@for_all_test_contexts
def test_ebe_monitor(test_context):
    line = xt.Line(elements=[xt.Multipole(knl=[0, 1.]),
                            xt.Drift(length=0.5),
                            xt.Multipole(knl=[0, -1]),
                            xt.Cavity(frequency=400e7, voltage=6e6),
                            xt.Drift(length=.5),
                            xt.Drift(length=0)])

    line.build_tracker(_context=test_context)

    particles = xp.Particles(x=[1e-3, -2e-3, 5e-3], y=[2e-3, -4e-3, 3e-3],
                            zeta=1e-2, p0c=7e12, mass0=xp.PROTON_MASS_EV,
                            _context=test_context)

    line.track(particles.copy(), turn_by_turn_monitor='ONE_TURN_EBE')

    mon = line.record_last_track

    for ii, ee in enumerate(line.elements):
        for tt, nn in particles.per_particle_vars:
            assert np.all(particles.to_dict()[nn] == getattr(mon, nn)[:, ii])
        ee.track(particles)
        particles.at_element += 1


@for_all_test_contexts
def test_cycle(test_context):
    d0 = xt.Drift()
    c0 = xt.Cavity()
    d1 = xt.Drift()
    r0 = xt.SRotation()
    particle_ref = xp.Particles(mass0=xp.PROTON_MASS_EV, gamma0=1.05)

    for collective in [True, False]:
        line = xt.Line(elements=[d0, c0, d1, r0])
        d1.iscollective = collective

        line.build_tracker(_context=test_context)
        line.particle_ref = particle_ref

        cline_name = line.copy()
        cline_index = line.copy()
        cline_name.build_tracker(_context=test_context)
        cline_index.build_tracker(_context=test_context)

        cline_name.cycle(name_first_element='e2')
        cline_index.cycle(index_first_element=2)

        assert cline_name.tracker is not None
        assert cline_index.tracker is not None

        for cline in [cline_index, cline_name]:
            assert cline.element_names[0] == 'e2'
            assert cline.element_names[1] == 'e3'
            assert cline.element_names[2] == 'e0'
            assert cline.element_names[3] == 'e1'

            assert isinstance(cline.elements[0], xt.Drift)
            assert isinstance(cline.elements[1], xt.SRotation)
            assert isinstance(cline.elements[2], xt.Drift)
            assert isinstance(cline.elements[3], xt.Cavity)

            assert cline.particle_ref.mass0 == xp.PROTON_MASS_EV
            assert cline.particle_ref.gamma0 == 1.05

@for_all_test_contexts
def test_synrad_configuration(test_context):
    for collective in [False, True]:
        elements = [xt.Multipole(knl=[1]) for _ in range(10)]
        if collective:
            elements[5].iscollective = True
            elements[5].move(_context=test_context)

        line = xt.Line(elements=elements)
        line.build_tracker(_context=test_context)

        line.configure_radiation(model='mean')
        for ee in line.elements:
            assert ee.radiation_flag == 1
        p = xp.Particles(x=[0.01, 0.02], _context=test_context)
        line.track(p)
        p.move(_context=xo.ContextCpu())
        assert np.all(p._rng_s1 + p._rng_s2 + p._rng_s3 + p._rng_s4 == 0)

        line.configure_radiation(model='quantum')
        for ee in line.elements:
            assert ee.radiation_flag == 2
        p = xp.Particles(x=[0.01, 0.02], _context=test_context)
        line.track(p)
        p.move(_context=xo.ContextCpu())
        assert np.all(p._rng_s1 + p._rng_s2 + p._rng_s3 + p._rng_s4 > 0)

        line.configure_radiation(model=None)
        for ee in line.elements:
            assert ee.radiation_flag == 0
        p = xp.Particles(x=[0.01, 0.02], _context=test_context)
        line.track(p)
        p.move(_context=xo.ContextCpu())
        assert np.all(p._rng_s1 + p._rng_s2 + p._rng_s3 + p._rng_s4 > 0)


@for_all_test_contexts
def test_partial_tracking(test_context):
    n_elem = 9
    elements = [ xt.Drift(length=1.) for _ in range(n_elem) ]
    line = xt.Line(elements=elements)
    line.build_tracker(_context=test_context)
    assert not line.iscollective
    particles_init = xp.Particles(_context=test_context,
        x=[1e-3, -2e-3, 5e-3], y=[2e-3, -4e-3, 3e-3],
        zeta=1e-2, p0c=7e12, mass0=xp.PROTON_MASS_EV,
        at_turn=0, at_element=0)

    _default_track(line, particles_init)
    _ele_start_until_end(line, particles_init)
    _ele_start_with_shift(line, particles_init)
    _ele_start_with_shift_more_turns(line, particles_init)
    _ele_stop_from_start(line, particles_init)
    _ele_start_to_ele_stop(line, particles_init)
    _ele_start_to_ele_stop_with_overflow(line, particles_init)


@for_all_test_contexts
def test_partial_tracking_with_collective(test_context):
    n_elem = 9
    elements = [xt.Drift(length=1., _context=test_context) for _ in range(n_elem)]
    # Make some elements collective
    elements[3].iscollective = True
    elements[7].iscollective = True
    line = xt.Line(elements=elements)
    line.build_tracker(_context=test_context)
    assert line.iscollective
    assert len(line.tracker._parts) == 5
    particles_init = xp.Particles(
            _context=test_context,
            x=[1e-3, -2e-3, 5e-3], y=[2e-3, -4e-3, 3e-3],
            zeta=1e-2, p0c=7e12, mass0=xp.PROTON_MASS_EV,
            at_turn=0, at_element=0)

    _default_track(line, particles_init)
    _ele_start_until_end(line, particles_init)
    _ele_start_with_shift(line, particles_init)
    _ele_start_with_shift_more_turns(line, particles_init)
    _ele_stop_from_start(line, particles_init)
    _ele_start_to_ele_stop(line, particles_init)
    _ele_start_to_ele_stop_with_overflow(line, particles_init)


# Track from the start until the end of the first, second, and tenth turn
def _default_track(line, particles_init):
    n_elem = len(line.element_names)
    for turns in [1, 2, 10]:
        expected_end_turn = turns
        expected_end_element = 0
        expected_num_monitor = expected_end_turn if expected_end_element==0 else expected_end_turn+1

        particles = particles_init.copy()
        line.track(particles, num_turns=turns, turn_by_turn_monitor=True)
        check, end_turn, end_element, end_s = _get_at_turn_element(particles)
        assert (check and end_turn==expected_end_turn and end_element==expected_end_element
                    and end_s==expected_end_element)
        assert line.record_last_track.x.shape == (len(particles.x), expected_num_monitor)


# Track, from any ele_start, until the end of the first, second, and tenth turn
def _ele_start_until_end(line, particles_init):
    n_elem = len(line.element_names)
    for turns in [1, 2, 10]:
        for start in range(n_elem):
            expected_end_turn = turns
            expected_end_element = 0
            expected_num_monitor = expected_end_turn if expected_end_element==0 else expected_end_turn+1

            particles = particles_init.copy()
            particles.at_element = start
            particles.s = start
            line.track(particles, num_turns=turns, ele_start=start, turn_by_turn_monitor=True)
            check, end_turn, end_element, end_s = _get_at_turn_element(particles)
            assert (check and end_turn==expected_end_turn and end_element==expected_end_element
                        and end_s==expected_end_element)
            assert line.record_last_track.x.shape==(len(particles.x), expected_num_monitor)


# Track, from any ele_start, any shifts that stay within the first turn
def _ele_start_with_shift(line, particles_init):
    n_elem = len(line.element_names)
    for start in range(n_elem):
        for shift in range(1,n_elem-start):
            expected_end_turn = 0
            expected_end_element = start+shift
            expected_num_monitor = expected_end_turn if expected_end_element==0 else expected_end_turn+1

            particles = particles_init.copy()
            particles.at_element = start
            particles.s = start
            line.track(particles, ele_start=start, num_elements=shift, turn_by_turn_monitor=True)
            check, end_turn, end_element, end_s = _get_at_turn_element(particles)
            assert (check and end_turn==expected_end_turn and end_element==expected_end_element
                        and end_s==expected_end_element)
            assert line.record_last_track.x.shape==(len(particles.x), expected_num_monitor)

# Track, from any ele_start, any shifts that are larger than one turn (up to 3 turns)
def _ele_start_with_shift_more_turns(line, particles_init):
    n_elem = len(line.element_names)
    for start in range(n_elem):
        for shift in range(n_elem-start, 3*n_elem+1):
            expected_end_turn = round(np.floor( (start+shift)/n_elem ))
            expected_end_element = start + shift - n_elem*expected_end_turn
            expected_num_monitor = expected_end_turn if expected_end_element==0 else expected_end_turn+1

            particles = particles_init.copy()
            particles.at_element = start
            particles.s = start
            line.track(particles, ele_start=start, num_elements=shift, turn_by_turn_monitor=True)
            check, end_turn, end_element, end_s = _get_at_turn_element(particles)
            assert (check and end_turn==expected_end_turn and end_element==expected_end_element
                        and end_s==expected_end_element)
            assert line.record_last_track.x.shape==(len(particles.x), expected_num_monitor)


# Track from the start until any ele_stop in the first, second, and tenth turn
def _ele_stop_from_start(line, particles_init):
    n_elem = len(line.element_names)
    for turns in [1, 2, 10]:
        for stop in range(1, n_elem):
            expected_end_turn = turns-1
            expected_end_element = stop
            expected_num_monitor = expected_end_turn if expected_end_element==0 else expected_end_turn+1

            particles = particles_init.copy()
            line.track(particles, num_turns=turns, ele_stop=stop, turn_by_turn_monitor=True)
            check, end_turn, end_element, end_s = _get_at_turn_element(particles)
            assert (check and end_turn==expected_end_turn and end_element==expected_end_element
                        and end_s==expected_end_element)
            assert line.record_last_track.x.shape==(len(particles.x), expected_num_monitor)


# Track from any ele_start until any ele_stop that is larger than ele_start
# for one, two, and ten turns
def _ele_start_to_ele_stop(line, particles_init):
    n_elem = len(line.element_names)
    for turns in [1, 2, 10]:
        for start in range(n_elem):
            for stop in range(start+1,n_elem):
                expected_end_turn = turns-1
                expected_end_element = stop
                expected_num_monitor = expected_end_turn if expected_end_element==0 else expected_end_turn+1

                particles = particles_init.copy()
                particles.at_element = start
                particles.s = start
                line.track(particles, num_turns=turns, ele_start=start, ele_stop=stop, turn_by_turn_monitor=True)
                check, end_turn, end_element, end_s = _get_at_turn_element(particles)
                assert (check and end_turn==expected_end_turn and end_element==expected_end_element
                            and end_s==expected_end_element)
                assert line.record_last_track.x.shape==(len(particles.x), expected_num_monitor)


# Track from any ele_start until any ele_stop that is smaller than or equal to ele_start (turn increses by one)
# for one, two, and ten turns
def _ele_start_to_ele_stop_with_overflow(line, particles_init):
    n_elem = len(line.element_names)
    for turns in [1, 2, 10]:
        for start in range(n_elem):
            for stop in range(start+1):
                expected_end_turn = turns
                expected_end_element = stop
                expected_num_monitor = expected_end_turn if expected_end_element==0 else expected_end_turn+1

                particles = particles_init.copy()
                particles.at_element = start
                particles.s = start
                line.track(particles, num_turns=turns, ele_start=start, ele_stop=stop, turn_by_turn_monitor=True)
                check, end_turn, end_element, end_s = _get_at_turn_element(particles)
                assert (check and end_turn==expected_end_turn and end_element==expected_end_element
                            and end_s==expected_end_element)
                assert line.record_last_track.x.shape==(len(particles.x), expected_num_monitor)


# Quick helper function to:
#   1) check that all survived particles are at the same element and turn
#   2) return that element and turn
def _get_at_turn_element(particles):
    part_cpu = particles.copy(_context=xo.ContextCpu())
    at_element = np.unique(part_cpu.at_element[part_cpu.state>0])
    at_turn = np.unique(part_cpu.at_turn[part_cpu.state>0])
    at_s = np.unique(part_cpu.s[part_cpu.state>0])
    all_together = len(at_turn)==1 and len(at_element)==1 and len(at_s)==1
    return all_together, at_turn[0], at_element[0], at_s[0]


def test_tracker_hashable_config():
    line = xt.Line([])
    line.build_tracker()
    line.config.TEST_FLAG_BOOL = True
    line.config.TEST_FLAG_INT = 42
    line.config.TEST_FLAG_FALSE = False
    line.config.ZZZ = 'lorem'
    line.config.AAA = 'ipsum'

    expected = (
        ('AAA', 'ipsum'),
        ('TEST_FLAG_BOOL', True),
        ('TEST_FLAG_INT', 42),
        ('XFIELDS_BB3D_NO_BEAMSTR', True), # active by default
        ('XFIELDS_BB3D_NO_BHABHA', True), # active by default
        ('XTRACK_GLOBAL_XY_LIMIT', 1.0), # active by default
        ('XTRACK_MULTIPOLE_NO_SYNRAD', True), # active by default
        ('ZZZ', 'lorem'),
    )

    assert line.tracker._hashable_config() == expected


def test_tracker_config_to_headers():
    line = xt.Line([])
    line.build_tracker()

    line.config.clear()
    line.config.TEST_FLAG_BOOL = True
    line.config.TEST_FLAG_INT = 42
    line.config.TEST_FLAG_FALSE = False
    line.config.ZZZ = 'lorem'
    line.config.AAA = 'ipsum'

    expected = [
        '#define TEST_FLAG_BOOL',
        '#define TEST_FLAG_INT 42',
        '#define ZZZ lorem',
        '#define AAA ipsum',
    ]

    assert set(line.tracker._config_to_headers()) == set(expected)


@for_all_test_contexts
def test_tracker_config(test_context):
    class TestElement(xt.BeamElement):
        _xofields = {
            'dummy': xo.Float64,
        }
        _extra_c_sources = ["""
            /*gpufun*/
            void TestElement_track_local_particle(
                    TestElementData el,
                    LocalParticle* part0)
            {
                //start_per_particle_block (part0->part)

                    #if TEST_FLAG == 2
                    LocalParticle_set_x(part, 7);
                    #endif

                    #ifdef TEST_FLAG_BOOL
                    LocalParticle_set_y(part, 42);
                    #endif

                //end_per_particle_block
            }
            """]

    test_element = TestElement(_context=test_context)
    line = xt.Line([test_element])
    line.build_tracker(_context=test_context)

    particles = xp.Particles(p0c=1e9, x=[0], y=[0], _context=test_context)

    p = particles.copy()
    line.config.TEST_FLAG = 2
    line.track(p)
    assert p.x[0] == 7.0
    assert p.y[0] == 0.0
    first_kernel, first_data = line.tracker.get_track_kernel_and_data_for_present_config()

    p = particles.copy()
    line.config.TEST_FLAG = False
    line.config.TEST_FLAG_BOOL = True
    line.track(p)
    assert p.x[0] == 0.0
    assert p.y[0] == 42.0
    current_kernel, current_data = line.tracker.get_track_kernel_and_data_for_present_config()
    assert current_kernel is not first_kernel
    assert current_data is not first_data

    line.config.TEST_FLAG = 2
    line.config.TEST_FLAG_BOOL = False
    assert len(line.tracker.track_kernel) == 3 # As line.track_kernel.keys() =
                                          # dict_keys([(), (('TEST_FLAG', 2),), (('TEST_FLAG_BOOL', True),)])
    current_kernel, current_data = line.tracker.get_track_kernel_and_data_for_present_config()
    assert current_kernel is first_kernel
    assert current_data is first_data

@for_all_test_contexts
@pytest.mark.parametrize('multiline', [False, True])
def test_optimize_for_tracking(test_context, multiline):

    if multiline:
        mline = xt.Environment.from_json(test_data_folder /
                         "hllhc15_collider/collider_00_from_mad.json")
        line = mline['lhcb1']
        line.particle_ref = xp.Particles(p0c=7000e9)
        line.vars['vrf400'] = 16
    else:
        fname_line_particles = test_data_folder / 'hllhc15_noerrors_nobb/line_and_particle.json'
        with open(fname_line_particles, 'r') as fid:
            input_data = json.load(fid)
        line = xt.Line.from_dict(input_data['line'])
        line.particle_ref = xp.Particles.from_dict(input_data['particle'])

    line.build_tracker(_context=test_context)

    particles = line.build_particles(
        x_norm=np.linspace(-2, 2, 1000), y_norm=0.1, delta=3e-4,
        nemitt_x=2.5e-6, nemitt_y=2.5e-6)

    p_no_optimized = particles.copy()
    p_optimized = particles.copy()

    num_turns = 10

    line.track(p_no_optimized, num_turns=num_turns, time=True)
    df_before_optimize = line.to_pandas()
    n_markers_before_optimize = (df_before_optimize.element_type == 'Marker').sum()
    assert n_markers_before_optimize > 4 # There are at least the IPs

    line.optimize_for_tracking(keep_markers=True)
    df_optimize_keep_markers = line.to_pandas()
    n_markers_optimize_keep = (df_optimize_keep_markers.element_type == 'Marker').sum()
    assert n_markers_optimize_keep == n_markers_before_optimize

    line.optimize_for_tracking(keep_markers=['ip1', 'ip5'])
    df_optimize_ip15 = line.to_pandas()
    n_markers_optimize_ip15 = (df_optimize_ip15.element_type == 'Marker').sum()
    assert n_markers_optimize_ip15 == 2

    line.optimize_for_tracking()

    assert type(line['mb.b10l3.b1..1']) is xt.SimpleThinBend
    assert type(line['mq.10l3.b1..1']) is xt.SimpleThinQuadrupole

    df_optimize = line.to_pandas()
    n_markers_optimize = (df_optimize.element_type == 'Marker').sum()
    assert n_markers_optimize == 0

    n_multipoles_before_optimize = (df_before_optimize.element_type == 'Multipole').sum()
    n_multipoles_optimize = (df_optimize.element_type == 'Multipole').sum()
    assert n_multipoles_before_optimize > n_multipoles_optimize

    n_drifts_before_optimize = (df_before_optimize.element_type == 'Drift').sum()
    n_drifts_optimize = (df_optimize.element_type == 'Drift').sum()
    assert n_drifts_before_optimize > n_drifts_optimize

    line.track(p_optimized, num_turns=num_turns, time=True)

    p_no_optimized.move(xo.context_default)
    p_optimized.move(xo.context_default)

    assert np.all(p_no_optimized.state == 1)
    assert np.all(p_optimized.state == 1)

    xo.assert_allclose(p_no_optimized.x, p_optimized.x, rtol=0, atol=1e-14)
    xo.assert_allclose(p_no_optimized.y, p_optimized.y, rtol=0, atol=1e-14)
    xo.assert_allclose(p_no_optimized.px, p_optimized.px, rtol=0, atol=1e-14)
    xo.assert_allclose(p_no_optimized.py, p_optimized.py, rtol=0, atol=1e-14)
    xo.assert_allclose(p_no_optimized.zeta, p_optimized.zeta, rtol=0, atol=1e-10)
    xo.assert_allclose(p_no_optimized.delta, p_optimized.delta, rtol=0, atol=1e-14)


@for_all_test_contexts
def test_backtrack_with_flag(test_context):

    line = xt.Line.from_json(test_data_folder /
                'hllhc15_noerrors_nobb/line_w_knobs_and_particle.json')
    line.build_tracker(_context=test_context)

    line.vars['on_crab1'] = -190
    line.vars['on_crab5'] = -190
    line.vars['on_x1'] = 130
    line.vars['on_x5'] = 130

    p = xp.Particles(_context=test_context,
        p0c=7000e9, x=1e-4, px=1e-6, y=2e-4, py=3e-6, zeta=0.01, delta=1e-4)

    line.track(p, turn_by_turn_monitor='ONE_TURN_EBE')
    mon_forward = line.record_last_track

    line.track(p, backtrack=True, turn_by_turn_monitor='ONE_TURN_EBE')
    mon_backtrack = line.record_last_track

    xo.assert_allclose(mon_forward.x, mon_backtrack.x, rtol=0, atol=1e-10)
    xo.assert_allclose(mon_forward.y, mon_backtrack.y, rtol=0, atol=1e-10)
    xo.assert_allclose(mon_forward.px, mon_backtrack.px, rtol=0, atol=1e-10)
    xo.assert_allclose(mon_forward.py, mon_backtrack.py, rtol=0, atol=1e-10)
    xo.assert_allclose(mon_forward.zeta, mon_backtrack.zeta, rtol=0, atol=1e-10)
    xo.assert_allclose(mon_forward.delta, mon_backtrack.delta, rtol=0, atol=1e-10)


@for_all_test_contexts
@pytest.mark.parametrize(
    'with_progress,turns',
    [(True, 300), (True, 317), (7, 523), (1, 21), (10, 10)]
)
@pytest.mark.parametrize('collective', [True, False], ids=['collective', 'non-collective'])
def test_tracking_with_progress(test_context, with_progress, turns, collective):
    elements = [xt.Drift(length=2, _context=test_context) for _ in range(5)]
    elements[3].iscollective = collective
    line = xt.Line(elements=elements)
    line.reset_s_at_end_turn = False

    particles = xp.Particles(x=[1e-3, 2e-3, 3e-3], p0c=7e12, _context=test_context)
    line.build_tracker(_context=test_context)
    line.track(particles, num_turns=turns, with_progress=with_progress)
    particles.move(xo.ContextCpu())

    assert np.all(particles.at_turn == turns)
    xo.assert_allclose(particles.s, 10 * turns, rtol=0, atol=1e-14)


@for_all_test_contexts
@pytest.mark.parametrize(
    'ele_start,ele_stop,expected_x',
    [
        (None, None, [0, 0.005, 0.010, 0.015, 0.020, 0.025]),
        (None, 3, [0, 0.005, 0.010, 0.015, 0.020, 0.023]),
        (2, None, [0, 0.003, 0.008, 0.013, 0.018, 0.023]),
        (2, 3, [0, 0.003, 0.008, 0.013, 0.018, 0.021]),
        (3, 2, [0, 0.002, 0.007, 0.012, 0.017, 0.022, 0.024]),
    ],
)
@pytest.mark.parametrize('with_progress', [False, True, 1, 2, 3])
def test_tbt_monitor_with_progress(test_context, ele_start, ele_stop, expected_x, with_progress):
    line = xt.Line(elements=[xt.Drift(length=1, _context=test_context)] * 5)
    line.build_tracker(_context=test_context)

    p = xt.Particles(px=0.001, _context=test_context)
    line.track(p, num_turns=5, turn_by_turn_monitor=True, with_progress=with_progress, ele_start=ele_start, ele_stop=ele_stop)
    p.move(_context=xo.context_default)

    monitor_recorded_x = line.record_last_track.x
    assert monitor_recorded_x.shape == (1, len(expected_x) - 1)

    recorded_x = np.concatenate([monitor_recorded_x[0], p.x])
    xo.assert_allclose(recorded_x, expected_x, atol=1e-16)


@pytest.fixture
def pimms_mad():
    pimms_path = test_data_folder / 'pimms/PIMMS.seq'
    mad = Madx(stdout=False)
    mad.option(echo=False)
    mad.call(str(pimms_path))
    mad.beam()
    mad.use('pimms')
    return mad


@for_all_test_contexts
@fix_random_seed(784239)
def test_track_log_and_merit_function(pimms_mad, test_context):
    line = xt.Line.from_madx_sequence(
        pimms_mad.sequence.pimms,
        deferred_expressions=True,
    )
    line.particle_ref = xt.Particles(
        q0=1,
        mass0=xt.PROTON_MASS_EV,
        kinetic_energy0=200e6, # eV
    )
    line.insert_element(
        name='septum_aperture',
        element=xt.LimitRect(min_x=-0.1, max_x=0.1, min_y=-0.1, max_y=0.1),
        at='extr_septum',
    )
    line['septum_aperture'].max_x = 0.035
    line.configure_bend_model(edge='full', core='adaptive', num_multipole_kicks=1)

    line.vars['kqf_common'] = 0
    line.vars['kqfa'] = line.vars['kqf_common']
    line.vars['kqfb'] = line.vars['kqf_common']

    line.vars['kqf_common'] = 2e-2
    line.vars['kqd'] = -2e-2

    line.build_tracker(_context=test_context)

    # Prepare a match on the tunes, dispersion at electrostatic septum, and
    # correct the chromaticity. Don't run it, instead we will simulate an
    # external optimizer.
    opt = line.match(
        solve=False,
        method='4d',
        vary=[
            xt.VaryList(['ksf', 'ksd'], step=1e-3),
            xt.VaryList(['kqfa', 'kqfb'], limits=(0, 1), step=1e-3, tag='qf'),
            xt.Vary('kqd', limits=(-1, 0), step=1e-3, tag='qd', weight=10),
        ],
        targets=[
            xt.TargetSet(dqx=-0.1, dqy=-0.1, tol=1e-3, tag="chrom"),
            xt.Target(dx=0, at='extr_septum', tol=1e-6),
            xt.TargetSet(qx=1.663, qy=1.72, tol=1e-6, tag="tunes"),
        ]
    )

    # Check that the merit function works correctly
    merit_function = opt.get_merit_function(return_scalar=True, check_limits=False)

    x_expected = [line.vars[kk]._value for kk in ['ksf', 'ksd', 'kqfa', 'kqfb', 'kqd']]
    x_expected[4] /= 10  # include the weight
    x_start = merit_function.get_x()
    xo.assert_allclose(x_start, x_expected, atol=1e-14)

    expected_limits = [
        [-1e200, 1e200],  # default
        [-1e200, 1e200],  # default
        [0, 1],
        [0, 1],
        [-0.1, 0],
    ]
    xo.assert_allclose(merit_function.get_x_limits(), expected_limits, atol=1e-14)

    # Below numbers obtained by first only matching the tunes, then the above
    x_optimized = [-1.40251213, 0.81823393, 0.31196667, 0.52478984, -0.052393429]
    merit_function.set_x(x_optimized)
    assert np.all(opt.target_status(ret=True)['tol_met'])

    # Now prepare to track and to log intensity and sextupole strength
    num_particles = 1000
    beam_intensity = 1e10  # p+

    # Generate particles with a gaussian distribution in normalized phase space,
    # and a gaussian momentum distribution (rms spread 5e-4)
    particles = line.build_particles(
        x_norm=np.random.normal(size=num_particles),
        px_norm=np.random.normal(size=num_particles),
        y_norm=np.random.normal(size=num_particles),
        py_norm=np.random.normal(size=num_particles),
        delta=5e-4 * np.random.normal(size=num_particles),
        method='4d',
        weight=beam_intensity / num_particles,
        nemitt_x=1.5e-6, nemitt_y=1e-6,
    )

    # Define time-dependent behaviour of the quadrupoles
    line.functions['fun_kqfa'] = xt.FunctionPieceWiseLinear(
        x=[0, 0.5e-3],
        y=[line.vv['kqfa'], 0.313],
    )
    line.vars['kqfa'] = line.functions['fun_kqfa'](line.vars['t_turn_s'])
    line.vars['kse2'] = 9

    kqfa_before = line.vv['kqfa']

    def measure_intensity(_, particles):
        ctx2np = particles._context.nparray_from_context_array
        state = ctx2np(particles.state)
        weight = ctx2np(particles.weight)
        mask_alive = state > 0
        return np.sum(weight[mask_alive])

    intensity_before = measure_intensity(None, particles)

    log = xt.Log('kqfa',  # vars to be logged
                 intensity=measure_intensity)  # user-defined function to be logged

    line.discard_tracker()
    line.build_tracker(_context=test_context)
    line.enable_time_dependent_vars = True
    num_turns = 1000
    line.track(particles=particles, num_turns=num_turns, log=log)

    intensity_after = measure_intensity(None, particles)

    # Check that kqfa is increasing linearly
    fit = np.polyfit(np.arange(num_turns), line.log_last_track['kqfa'], 1, full=True)
    (slope, _), residual, _, _, _ = fit
    assert slope > 0
    assert residual < 1e-28
    xo.assert_allclose(line.log_last_track['kqfa'][0], kqfa_before, atol=1e-14, rtol=0)
    xo.assert_allclose(line.log_last_track['kqfa'][-1], line.vv['kqfa'], atol=1e-14, rtol=0)

    # Check that intensity is decreasing
    intensity = np.array(line.log_last_track['intensity'])
    assert np.all(intensity[:-1] - intensity[1:] >= 0)
    xo.assert_allclose(intensity[0], intensity_before, atol=1e-14, rtol=0)
    # The last log point is from the beginning of the last turn:
    xo.assert_allclose(intensity[-1], intensity_after, atol=0, rtol=1e-2)


@for_all_test_contexts
def test_init_io_buffer(test_context):
    class TestElementRecord(xo.HybridClass):
        _xofields = {
            '_index': xt.RecordIndex,
            'record_field': xo.Int64[:],
            'record_at_element': xo.Int64[:],
        }

    class TestElement(xt.BeamElement):
        _xofields={
            'element_field': xo.Int64,
        }

        _internal_record_class = TestElementRecord

        _extra_c_sources = [
            r'''
            /*gpufun*/
            void TestElement_track_local_particle(TestElementData el, LocalParticle* part0){
                // Extract the record and record_index
                TestElementRecordData record = TestElementData_getp_internal_record(el, part0);
                RecordIndex record_index = NULL;
                if (record){
                    record_index = TestElementRecordData_getp__index(record);
                }

                int64_t elem_field = TestElementData_get_element_field(el);

                //start_per_particle_block (part0->part)
                    if (record) {  // Record exists
                        // Get a slot in the record (this is thread safe)
                        int64_t i_slot = RecordIndex_get_slot(record_index);

                        if (i_slot>=0) {  // Slot available
                            TestElementRecordData_set_record_field(
                                record,
                                i_slot,
                                elem_field
                            );
                            TestElementRecordData_set_record_at_element(
                                record,
                                i_slot,
                                LocalParticle_get_at_element(part)
                            );
                        }
                    }
                //end_per_particle_block
            }
            '''
        ]

    line = xt.Line(elements=[
        TestElement(element_field=3),
        TestElement(element_field=4),
    ])
    line.build_tracker(_context=test_context)

    record = line.start_internal_logging_for_elements_of_type(
        TestElement,
        capacity=1000,
    )

    num_turns = 100

    part = xp.Particles(_context=test_context, x=[1e-3, 2e-3, 3e-3])
    line.track(part, num_turns=num_turns)

    num_recorded = record._index.num_recorded
    num_particles = len(part.x)

    record.move(_context=xo.ContextCpu())

    assert num_recorded == (2 * num_particles * num_turns)
    assert np.sum(record.record_field == 3) == num_particles * num_turns
    assert np.sum(record.record_field == 4) == num_particles * num_turns
    assert np.all(record.record_field[:num_recorded][
                    record.record_at_element[:num_recorded] == 0] == 3)
    assert np.all(record.record_field[:num_recorded][
                    record.record_at_element[:num_recorded] == 1] == 4)
    assert np.all(record.record_field[num_recorded:] == 0)

    # Now we stop logging and manually reset to mimic the situation where the
    # record is manually flushed.
    line.stop_internal_logging_for_elements_of_type(TestElement)
    line.tracker._init_io_buffer()

    assert line.tracker.io_buffer is not None



import xtrack as xt
import time
import numpy as np

# Load a line and build tracker
line = xt.Line.from_json(
    '../../test_data/hllhc15_thick/lhc_thick_with_knobs.json')
line.particle_ref = xt.Particles(mass0=xt.PROTON_MASS_EV, q0=1, energy0=7e12)
line.vars['vrf400'] = 16
line.build_tracker()

# To store kernels
tw0_6d = line.twiss()
tw0_4d = line.twiss4d()

# Time twiss 4d
start = time.time()
tw4d = line.twiss4d()
end = time.time()
print(f'Time for twiss4d: {end-start}')

p = line.build_particles(x=np.linspace(-1e-3, 1e-3, 1000))
line.track(p, num_turns=20, with_progress=1, time=True)
line.time_last_track



# copyright ############################### #
# This file is part of the Xtrack Package.  #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import xobjects as xo
import xtrack as xt

from ..base_element import BeamElement
from ..general import _pkg_root

def _monitor_init(
    self,
    _context=None,
    _buffer=None,
    _offset=None,
    _xobject=None,
    start_at_turn=None,
    stop_at_turn=None,
    n_repetitions=None,
    repetition_period=None,
    num_particles=None,
    particle_id_range=None,
    auto_to_numpy=True,
):

    '''
    Beam element logging the coordinates of the particles passing through it.

    Parameters
    ----------
    start_at_turn: int
        Turn at which the monitor starts logging the particles coordinates.
    stop_at_turn: int
        Turn at which the monitor stops logging the particles coordinates.
    n_repetitions: int
        Number of times the monitor repeats the logging of the particles.
    repetition_period: int
        Period in number of turns for the repetition of the logging of the particles.
    num_particles: int
        Number of particles to be logged.
    particle_id_range: tuple of int
        Range of particle ids to be logged.
    auto_to_numpy: bool
        If True, the data is automatically converted to numpy arrays when
        accessed.

    '''

    if _xobject is not None:
        self.xoinitialize(_xobject=_xobject)
    else:
        if particle_id_range is not None:
            assert num_particles is None
            part_id_start = particle_id_range[0]
            part_id_end = particle_id_range[1]
        else:
            assert num_particles is not None
            part_id_start = 0
            part_id_end = num_particles

        n_part_ids = part_id_end - part_id_start
        assert n_part_ids >= 0

        n_turns = int(stop_at_turn) - int(start_at_turn)

        if repetition_period is not None:
            assert n_repetitions is not None

        if n_repetitions is not None:
            assert repetition_period is not None

        if repetition_period is None:
            repetition_period = -1
            n_repetitions = 1

        n_records = n_turns * n_part_ids * n_repetitions

        data_init = {nn: n_records for tt, nn in
                        self._ParticlesClass.per_particle_vars}

        self.xoinitialize(
            _context=_context,
            _buffer=_buffer,
            _offset=_offset,
            start_at_turn=start_at_turn,
            stop_at_turn=stop_at_turn,
            part_id_start=part_id_start,
            part_id_end=part_id_end,
            n_records=n_records,
            n_repetitions=n_repetitions,
            repetition_period=repetition_period,
            data=data_init,
        )

        self._dressed_data = self._ParticlesClass(_xobject=self._xobject.data,
                                                  _no_reorganize=True)
        self.auto_to_numpy = auto_to_numpy

        with self.data._bypass_linked_vars():
            for tt, nn in self._ParticlesClass.per_particle_vars:
                getattr(self.data, nn)[:] = 0

def monitor_from_dict(cls, dct, **kwargs):
    xobj = cls._XoStruct(**dct, **kwargs)
    return cls(_xobject=xobj)

def auto_to_numpy(self):
    return self.flag_auto_to_numpy != 0

def set_auto_to_numpy(self, flag):
    if flag:
        self.flag_auto_to_numpy = 1
    else:
        self.flag_auto_to_numpy = 0

class _FieldOfMonitor:
    def __init__(self, name):
        self.name = name

    def __get__(self, container, ContainerType=None):
        vv = getattr(container.data, self.name)
        if container.auto_to_numpy:
            ctx = container._buffer.context
            vv = ctx.nparray_from_context_array(vv)

        n_cols = container.stop_at_turn - container.start_at_turn

        if container.n_repetitions == 1:
            n_rows = container.n_records // n_cols
            return vv.reshape(n_rows, n_cols)
        else:
            n_rows = container.n_records // n_cols // container.n_repetitions
            #return vv.reshape(container.n_repetitions, n_cols, n_rows)
            return vv.reshape(container.n_repetitions, n_rows, n_cols)


def _monitor_get_backtrack_element(
                    self, _context=None, _buffer=None, _offset=None):

    return xt.Marker(_context=_context, _buffer=_buffer, _offset=_offset)


class ParticlesMonitor(BeamElement):

    _xofields = {
        "start_at_turn": xo.Int64,
        "stop_at_turn": xo.Int64,
        'part_id_start': xo.Int64,
        'part_id_end': xo.Int64,
        'ebe_mode': xo.Int64,
        "n_records": xo.Int64,
        "n_repetitions": xo.Int64,
        "repetition_period": xo.Int64,
        "flag_auto_to_numpy": xo.Int64,
        "data": xt.Particles,
    }

    _extra_c_sources = [
        _pkg_root.joinpath("monitors/particles_monitor.h")
    ]

    behaves_like_drift = True
    has_backtrack = True
    allow_loss_refinement = True
    _ParticlesClass = xt.Particles


ParticlesMonitor.__init__ = _monitor_init
ParticlesMonitor.get_backtrack_element = _monitor_get_backtrack_element
ParticlesMonitor.from_dict = classmethod(monitor_from_dict)

ParticlesMonitor.auto_to_numpy = property(auto_to_numpy, set_auto_to_numpy)

per_particle_vars = xt.Particles.per_particle_vars
for tt, nn in per_particle_vars:
    setattr(ParticlesMonitor, nn, _FieldOfMonitor(name=nn))

for nn in ['pzeta', 'kin_px', 'kin_py', 'kin_ps', 'kin_xprime', 'kin_yprime']:
    setattr(ParticlesMonitor, nn, _FieldOfMonitor(name=nn))







import json
import numpy as np
import xtrack as xt


case_name = 'fcc-ee'
filename = 'line_no_radiation.json'

with open(filename, 'r') as f:
    line = xt.Line.from_dict(json.load(f))

#line['qc1l1.1..1'].knl[0] += 1e-5
#line['qc1l1.1..1'].ksl[0] += 1e-6/5
line['qc1l1.1..1'].ksl[1] = 1e-4

line.build_tracker()

# Initial twiss (no radiation)
line.configure_radiation(model=None)
tw_no_rad = line.twiss(method='4d', freeze_longitudinal=True)

# Enable radiation
line.configure_radiation(model='mean')
# - Set cavity lags to compensate energy loss
# - Taper magnet strengths
line.compensate_radiation_energy_loss(record_iterations=True)

import matplotlib.pyplot as plt
plt.close('all')

tw = line.twiss(eneloss_and_damping=True)

p0corr = 1 + line.delta_taper

plt.figure(1, figsize=(6.4*1.3, 4.8))

betx_beat = tw.betx*p0corr/tw_no_rad.betx-1
bety_beat = tw.bety*p0corr/tw_no_rad.bety-1
max_betx_beat = np.max(np.abs(betx_beat))
max_bety_beat = np.max(np.abs(bety_beat))
spx = plt.subplot(2,1,1)
sp0 = spx
plt.title(f'error on Qx: {abs(tw.qx - tw_no_rad.qx):.2e}     '
            r'$(\Delta \beta_x / \beta_x)_{max}$ = '
            f'{max_betx_beat:.2e}')
plt.plot(tw.s, betx_beat)
plt.ylabel(r'$\Delta \beta_x / \beta_x$')
plt.ylim(np.max([0.01, 1.1 * max_betx_beat])*np.array([-1, 1]))
plt.xlim([0, tw.s[-1]])

plt.subplot(2,1,2, sharex=sp0)
plt.title(f'error on Qy: {abs(tw.qy - tw_no_rad.qy):.2e}     '
            r'$(\Delta \beta_y / \beta_y)_{max}$ = '
            f'{max_bety_beat:.2e}')
plt.plot(tw.s, bety_beat)
plt.ylabel(r'$\Delta \beta_y / \beta_y$')
plt.ylim(np.max([0.01, 1.1 * max_bety_beat])*np.array([-1, 1]))
plt.xlabel('s [m]')

plt.subplots_adjust(hspace=0.35, top=.85)


plt.figure(11, figsize=(6.4*1.3, 4.8))

spx = plt.subplot(2,1,1, sharex=sp0)
plt.plot(tw.s, tw.betx, label='tapered')
plt.plot(tw.s, tw_no_rad.betx, '--', label='no radiation')
plt.ylabel(r'$\beta_{x}}$ [m]')
plt.xlim([0, tw.s[-1]])
plt.xlabel('s [m]')
plt.legend(loc='upper right')

plt.subplot(2,1,2, sharex=spx)
plt.plot(tw.s, tw.bety)
plt.plot(tw.s, tw_no_rad.bety, '--')
plt.ylabel(r'$\beta_{y}$ [m]')
plt.xlabel('s [m]')

plt.subplots_adjust(hspace=0.35, top=.85)


plt.figure(2, figsize=(6.4*1.3, 4.8))

bety1_beat = tw.bety1*p0corr/tw_no_rad.bety1-1
betx2_beat = tw.betx2*p0corr/tw_no_rad.betx2-1
max_bety1_beat = np.max(np.abs(bety1_beat))
max_betx2_beat = np.max(np.abs(betx2_beat))
spx = plt.subplot(2,1,1, sharex=sp0)
plt.title(f'error on Qx: {abs(tw.qx - tw_no_rad.qx):.2e}     '
            r'$(\Delta \beta_{y1} / \beta_{y1})_{max}$ = '
            f'{max_bety1_beat:.2e}')
plt.plot(tw.s, bety1_beat)
plt.ylabel(r'$\Delta \beta_{y1} / \beta_{y1}$')
plt.ylim(np.max([0.01, 1.1 * max_bety1_beat])*np.array([-1, 1]))
plt.xlim([0, tw.s[-1]])

plt.subplot(2,1,2, sharex=spx)
plt.title(f'error on Qy: {abs(tw.qy - tw_no_rad.qy):.2e}     '
            r'$(\Delta \beta_{x2} / \beta_{x2})_{max}$ = '
            f'{max_betx2_beat:.2e}')
plt.plot(tw.s, betx2_beat)
plt.ylabel(r'$\Delta \beta_{x2} / \beta_{x2}$')
plt.ylim(np.max([0.01, 1.1 * max_betx2_beat])*np.array([-1, 1]))
plt.xlabel('s [m]')

plt.subplots_adjust(hspace=0.35, top=.85)



plt.figure(22, figsize=(6.4*1.3, 4.8))

spx = plt.subplot(2,1,1, sharex=sp0)
plt.plot(tw.s, tw.bety1, label='tapered')
plt.plot(tw.s, tw_no_rad.bety1, '--', label='no radiation')
plt.ylabel(r'$\beta_{y1}}$ [m]')
plt.xlim([0, tw.s[-1]])
plt.xlabel('s [m]')
plt.legend(loc='upper right')

plt.subplot(2,1,2, sharex=spx)
plt.plot(tw.s, tw.betx2)
plt.plot(tw.s, tw_no_rad.betx2, '--')
plt.ylabel(r'$\beta_{x2}$ [m]')
plt.xlabel('s [m]')

plt.subplots_adjust(hspace=0.35, top=.85)


plt.figure(33, figsize=(6.4*1.3, 4.8))

spx = plt.subplot(2,1,1, sharex=sp0)
plt.plot(tw.s, tw.x, label='tapered')
plt.plot(tw.s, tw_no_rad.x, '--', label='no radiation')
plt.ylabel(r'x [m]')
plt.xlim([0, tw.s[-1]])
plt.xlabel('s [m]')
plt.legend(loc='upper right')

plt.subplot(2,1,2, sharex=spx)
plt.plot(tw.s, tw.y)
plt.plot(tw.s, tw_no_rad.y, '--')
plt.ylabel(r'y [m]')
plt.xlabel('s [m]')

plt.subplots_adjust(hspace=0.35, top=.85)




plt.figure(100)
for i_iter, mon in enumerate(line._tapering_iterations):
    plt.plot(mon.s.T, mon.delta.T,
                label=f'iter {i_iter} - Ene. loss: {-mon.delta[-1, -1]*100:.2f} %')
    plt.legend(loc='lower left')
    plt.grid(True)
    plt.ylim(top=0.01)
    plt.xlim(left=0, right=mon.s[-1, -1])
    plt.xlabel('s [m]')
    plt.ylabel(r'$\delta$')

plt.show()



import xtrack as xt

# --- Parameters
seq         = 'lhcb1'
ip_name     = 'ip1'
s_marker    = f'e.ds.l{ip_name[-1]}.b1'
e_marker    = f's.ds.r{ip_name[-1]}.b1'
#-------------------------------------


collider_file = '../../test_data/hllhc15_collider/collider_00_from_mad.json'


# Load the machine and select line
collider= xt.Environment.from_json(collider_file)
collider.vars['test_vars'] = 3.1416
line   = collider[seq]
line_sel    = line.select(s_marker,e_marker)

assert line_sel.element_dict is line.element_dict
assert line.get('ip1') is line_sel.get('ip1')

line_sel['aaa'] = 1e-6
assert line_sel['aaa'] == 1e-6
assert line['aaa'] == 1e-6

line_sel.ref['mcbch.7r1.b1'].knl[0] += line.ref['aaa']
assert (str(line.ref['mcbch.7r1.b1'].knl[0]._expr)
        == "((-vars['acbch7.r1b1']) + vars['aaa'])")
assert (str(line_sel.ref['mcbch.7r1.b1'].knl[0]._expr)
        == "((-vars['acbch7.r1b1']) + vars['aaa'])")
assert line_sel.get('mcbch.7r1.b1').knl[0] == 1e-6
assert line.get('mcbch.7r1.b1').knl[0] == 1e-6




import xtrack as xt

env = xt.Environment()
env.call('lhc_seq.py')

env.lhcb1.particle_ref = xt.Particles(p0c=0.45e12)
env.lhcb2.particle_ref = xt.Particles(p0c=0.45e12)

env.vars.load_madx('../../test_data/lhc_2024/injection_optics.madx')

ttvars = env.vars.get_table()
for nn in ttvars.rows['on_.*'].name:
    env[nn] = 0

tw1 = env.lhcb1.twiss4d()
tw2 = env.lhcb2.twiss4d(reverse=True)


import xtrack as xt

# Create an environment
env = xt.Environment()

# Create a line with two quadrupoles and a marker
line = env.new_line(name='myline', components=[
    env.new('q0', xt.Quadrupole, length=2.0, at=10.),
    env.new('q1', xt.Quadrupole, length=2.0, at=20.),
    env.new('m0', xt.Marker, at=50),
    ])

tt0 = line.get_table()
tt0.show(cols=['s_start', 's_center', 's_end'])
# is:
# drift_1                0           4.5             9
# q0                     9            10            11
# drift_2               11            15            19
# q1                    19            20            21
# drift_3               21          35.5            50
# m0                    50            50            50

# Create a set of new elements to be placed
env.new('s1', xt.Sextupole, length=2., k2=0.2)
env.new('s2', xt.Sextupole, length=2., k2=-0.2)
env.new('m1', xt.Marker)
env.new('m2', xt.Marker)
env.new('m3', xt.Marker)

subline = env.new_line(components=[
    env.place('s1', at=1.0),
    env.place('s2', at=5.0),
    env.place('m1', at='s1@start'),
    env.place('m2', at='s2@start')])

# Insert the new elements in the line
line.insert([
    env.place(subline, anchor='start', at=1., from_='q1@end'),
    env.place('m3', at=30.)])

tt = line.get_table()
tt.show(cols=['s_start', 's_center', 's_end'])
# is
# name             s_start      s_center         s_end
# drift_1                0           4.5             9
# q0                     9            10            11
# drift_2               11            15            19
# q1                    19            20            21
# drift_3..0            21          21.5            22
# m1                    22            22            22
# s1                    22            23            24
# drift_4               24            25            26
# m2                    26            26            26
# s2                    26            27            28
# drift_3..4            28            29            30
# m3                    30            30            30
# drift_3..5            30            40            50
# m0                    50            50            50
# _end_point            50            50            50


import xtrack as xt
import numpy as np
from xobjects.test_helpers import for_all_test_contexts
import pytest

assert_allclose = np.testing.assert_allclose

def _make_line_no_expressions():
    bend = xt.Bend(
        k0=0.4, angle=0.3, length=1,
        shift_x=1e-3, shift_y=2e-3, shift_s=2e-3, rot_s_rad=0.2,
        k1=0.1,
        knl=[0.7, 0.8, 0.9, 1.0], ksl=[0.1, 0.2, 0.3, 0.4])
    quad = xt.Quadrupole(k1=0.1, k1s=0.2,
        length=0.5, shift_x=2e-3, shift_y=1e-3, shift_s=-1e-3, rot_s_rad=0.1)
    sext = xt.Sextupole(k2=0.1, k2s=0.2,
        length=0.3, shift_x=3e-3, shift_y=4e-3, shift_s=-2e-3, rot_s_rad=0.3)
    octu = xt.Octupole(k3=0.1, k3s=0.2,
        length=0.4, shift_x=5e-3, shift_y=6e-3, shift_s=-1e-3, rot_s_rad=0.4)
    mult = xt.Multipole(knl=[0.7, 0.8, 0.9, 1.0], ksl=[0.1, 0.2, 0.3, 0.4],
        length=0.4, shift_x=5e-3, shift_y=6e-3, shift_s=-1e-3, rot_s_rad=0.7,
        hxl=0.1)
    drift = xt.Drift(length=5.0)

    line = xt.Line(elements=[bend, quad, sext, octu, mult, drift, xt.Replica(parent_name='e0')])

    return line

def _make_line_with_expressions():
    bend = xt.Bend(length=999.)
    quad = xt.Quadrupole(length=999.)
    sext = xt.Sextupole(length=999.)
    octu = xt.Octupole(length=999.)
    mult = xt.Multipole(length=999, knl=[999.]*4, ksl=[999.]*4)
    drift = xt.Drift()

    line = xt.Line(elements=[bend, quad, sext, octu, mult, drift, xt.Replica(parent_name='e0')])

    line.vars['k0_bend'] = 999.
    line.vars['angle_bend'] = 999.
    line.vars['length_bend'] = 999.
    line.vars['shift_x_bend'] = 999.
    line.vars['shift_y_bend'] = 999.
    line.vars['shift_s_bend'] = 999.
    line.vars['rot_s_rad_bend'] = 999.
    line.vars['k1_bend'] = 999.
    line.vars['knl_bend_0'] = 999.
    line.vars['knl_bend_1'] = 999.
    line.vars['knl_bend_2'] = 999.
    line.vars['knl_bend_3'] = 999.
    line.vars['ksl_bend_0'] = 999.
    line.vars['ksl_bend_1'] = 999.
    line.vars['ksl_bend_2'] = 999.
    line.vars['ksl_bend_3'] = 999.

    line.vars['k1_quad'] = 999.
    line.vars['k1s_quad'] = 999.
    line.vars['length_quad'] = 999.
    line.vars['shift_x_quad'] = 999.
    line.vars['shift_y_quad'] = 999.
    line.vars['shift_s_quad'] = 999.
    line.vars['rot_s_rad_quad'] = 999.

    line.vars['k2_sext'] = 999.
    line.vars['k2s_sext'] = 999.
    line.vars['length_sext'] = 999.
    line.vars['shift_x_sext'] = 999.
    line.vars['shift_y_sext'] = 999.
    line.vars['shift_s_sext'] = 999.
    line.vars['rot_s_rad_sext'] = 999.

    line.vars['k3_octu'] = 999.
    line.vars['k3s_octu'] = 999.
    line.vars['length_octu'] = 999.
    line.vars['shift_x_octu'] = 999.
    line.vars['shift_y_octu'] = 999.
    line.vars['shift_s_octu'] = 999.
    line.vars['rot_s_rad_octu'] = 999.

    line.vars['knl_mult_0'] = 999.
    line.vars['knl_mult_1'] = 999.
    line.vars['knl_mult_2'] = 999.
    line.vars['knl_mult_3'] = 999.
    line.vars['ksl_mult_0'] = 999.
    line.vars['ksl_mult_1'] = 999.
    line.vars['ksl_mult_2'] = 999.
    line.vars['ksl_mult_3'] = 999.
    line.vars['length_mult'] = 999.
    line.vars['shift_x_mult'] = 999.
    line.vars['shift_y_mult'] = 999.
    line.vars['shift_s_mult'] = 999.
    line.vars['rot_s_rad_mult'] = 999.
    line.vars['hxl_mult'] = 999.

    line.vars['length_drift'] = 999.

    line.element_refs['e0'].k0 = line.vars['k0_bend']
    line.element_refs['e0'].angle = line.vars['angle_bend']
    line.element_refs['e0'].length = line.vars['length_bend']
    line.element_refs['e0'].shift_x = line.vars['shift_x_bend']
    line.element_refs['e0'].shift_y = line.vars['shift_y_bend']
    line.element_refs['e0'].shift_s = line.vars['shift_s_bend']
    line.element_refs['e0'].rot_s_rad = line.vars['rot_s_rad_bend']
    line.element_refs['e0'].k1 = line.vars['k1_bend']
    line.element_refs['e0'].knl[0] = line.vars['knl_bend_0']
    line.element_refs['e0'].knl[1] = line.vars['knl_bend_1']
    line.element_refs['e0'].knl[2] = line.vars['knl_bend_2']
    line.element_refs['e0'].knl[3] = line.vars['knl_bend_3']
    line.element_refs['e0'].ksl[0] = line.vars['ksl_bend_0']
    line.element_refs['e0'].ksl[1] = line.vars['ksl_bend_1']
    line.element_refs['e0'].ksl[2] = line.vars['ksl_bend_2']
    line.element_refs['e0'].ksl[3] = line.vars['ksl_bend_3']

    line.element_refs['e1'].k1 = line.vars['k1_quad']
    line.element_refs['e1'].k1s = line.vars['k1s_quad']
    line.element_refs['e1'].length = line.vars['length_quad']
    line.element_refs['e1'].shift_x = line.vars['shift_x_quad']
    line.element_refs['e1'].shift_y = line.vars['shift_y_quad']
    line.element_refs['e1'].shift_s = line.vars['shift_s_quad']
    line.element_refs['e1'].rot_s_rad = line.vars['rot_s_rad_quad']

    line.element_refs['e2'].k2 = line.vars['k2_sext']
    line.element_refs['e2'].k2s = line.vars['k2s_sext']
    line.element_refs['e2'].length = line.vars['length_sext']
    line.element_refs['e2'].shift_x = line.vars['shift_x_sext']
    line.element_refs['e2'].shift_y = line.vars['shift_y_sext']
    line.element_refs['e2'].shift_s = line.vars['shift_s_sext']
    line.element_refs['e2'].rot_s_rad = line.vars['rot_s_rad_sext']

    line.element_refs['e3'].k3 = line.vars['k3_octu']
    line.element_refs['e3'].k3s = line.vars['k3s_octu']
    line.element_refs['e3'].length = line.vars['length_octu']
    line.element_refs['e3'].shift_x = line.vars['shift_x_octu']
    line.element_refs['e3'].shift_y = line.vars['shift_y_octu']
    line.element_refs['e3'].shift_s = line.vars['shift_s_octu']
    line.element_refs['e3'].rot_s_rad = line.vars['rot_s_rad_octu']

    line.element_refs['e4'].knl[0] = line.vars['knl_mult_0']
    line.element_refs['e4'].knl[1] = line.vars['knl_mult_1']
    line.element_refs['e4'].knl[2] = line.vars['knl_mult_2']
    line.element_refs['e4'].knl[3] = line.vars['knl_mult_3']
    line.element_refs['e4'].ksl[0] = line.vars['ksl_mult_0']
    line.element_refs['e4'].ksl[1] = line.vars['ksl_mult_1']
    line.element_refs['e4'].ksl[2] = line.vars['ksl_mult_2']
    line.element_refs['e4'].ksl[3] = line.vars['ksl_mult_3']
    line.element_refs['e4'].length = line.vars['length_mult']
    line.element_refs['e4'].shift_x = line.vars['shift_x_mult']
    line.element_refs['e4'].shift_y = line.vars['shift_y_mult']
    line.element_refs['e4'].shift_s = line.vars['shift_s_mult']
    line.element_refs['e4'].rot_s_rad = line.vars['rot_s_rad_mult']
    line.element_refs['e4'].hxl = line.vars['hxl_mult']

    line.element_refs['e5'].length = line.vars['length_drift']

    return line

def _set_vars(line):
    line.vars['k0_bend'] = 0.4
    line.vars['angle_bend'] = 0.3
    line.vars['length_bend'] = 1
    line.vars['shift_x_bend'] = 1e-3
    line.vars['shift_y_bend'] = 2e-3
    line.vars['shift_s_bend'] = 2e-3
    line.vars['rot_s_rad_bend'] = 0.2
    line.vars['k1_bend'] = 0.1
    line.vars['knl_bend_0'] = 0.7
    line.vars['knl_bend_1'] = 0.8
    line.vars['knl_bend_2'] = 0.9
    line.vars['knl_bend_3'] = 1.0
    line.vars['ksl_bend_0'] = 0.1
    line.vars['ksl_bend_1'] = 0.2
    line.vars['ksl_bend_2'] = 0.3
    line.vars['ksl_bend_3'] = 0.4

    line.vars['k1_quad'] = 0.1
    line.vars['k1s_quad'] = 0.2
    line.vars['length_quad'] = 0.5
    line.vars['shift_x_quad'] = 2e-3
    line.vars['shift_y_quad'] = 1e-3
    line.vars['shift_s_quad'] = -1e-3
    line.vars['rot_s_rad_quad'] = 0.1

    line.vars['k2_sext'] = 0.1
    line.vars['k2s_sext'] = 0.2
    line.vars['length_sext'] = 0.3
    line.vars['shift_x_sext'] = 3e-3
    line.vars['shift_y_sext'] = 4e-3
    line.vars['shift_s_sext'] = -2e-3
    line.vars['rot_s_rad_sext'] = 0.3

    line.vars['k3_octu'] = 0.1
    line.vars['k3s_octu'] = 0.2
    line.vars['length_octu'] = 0.4
    line.vars['shift_x_octu'] = 5e-3
    line.vars['shift_y_octu'] = 6e-3
    line.vars['shift_s_octu'] = -1e-3
    line.vars['rot_s_rad_octu'] = 0.4

    line.vars['knl_mult_0'] = 0.7
    line.vars['knl_mult_1'] = 0.8
    line.vars['knl_mult_2'] = 0.9
    line.vars['knl_mult_3'] = 1.0
    line.vars['ksl_mult_0'] = 0.1
    line.vars['ksl_mult_1'] = 0.2
    line.vars['ksl_mult_2'] = 0.3
    line.vars['ksl_mult_3'] = 0.4
    line.vars['length_mult'] = 0.4
    line.vars['shift_x_mult'] = 5e-3
    line.vars['shift_y_mult'] = 6e-3
    line.vars['shift_s_mult'] = -1e-3
    line.vars['rot_s_rad_mult'] = 0.7
    line.vars['hxl_mult'] = 0.1

    line.vars['length_drift'] = 5.0


@for_all_test_contexts
@pytest.mark.parametrize(
    "check_expr",
    [False, True],
    ids=['no_expr', 'with_expr'],
)
@pytest.mark.parametrize(
    "use_copy",
    [False, True],
    ids=['no_copy', 'with_copy'],
)
def test_attr_replicas(test_context, check_expr, use_copy):

    if check_expr:
        line = _make_line_with_expressions()
    else:
        line = _make_line_no_expressions()

    line.build_tracker(_context=test_context)

    if use_copy:
        line = line.copy()
        line.build_tracker(_context=test_context)

    if check_expr:
        _set_vars(line)

    tt = line.get_table(attr=True)

    assert_allclose(tt['s', -1], 8.2, rtol=0, atol=1e-14)

    assert tt['element_type','e0'] == 'Bend'
    assert tt['isreplica', 'e0'] == False
    assert tt['parent_name', 'e0'] is None
    assert tt['isthick', 'e0'] == True
    assert tt['iscollective', 'e0'] == False
    assert_allclose(tt['s', 'e0'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['length', 'e0'], 1, rtol=0, atol=1e-14)
    assert_allclose(tt['angle_rad', 'e0'], 0.3, rtol=0, atol=1e-14)
    assert_allclose(tt['rot_s_rad', 'e0'], 0.2, rtol=0, atol=1e-14)
    assert_allclose(tt['shift_x', 'e0'], 1e-3, rtol=0, atol=1e-14)
    assert_allclose(tt['shift_y', 'e0'], 2e-3, rtol=0, atol=1e-14)
    assert_allclose(tt['shift_s', 'e0'], 2e-3, rtol=0, atol=1e-14)
    assert_allclose(tt['k0l', 'e0'], 0.4 * 1 + 0.7, rtol=0, atol=1e-14)
    assert_allclose(tt['k1l', 'e0'], 0.1 * 1 + 0.8, rtol=0, atol=1e-14)
    assert_allclose(tt['k2l', 'e0'], 0.9, rtol=0, atol=1e-14)
    assert_allclose(tt['k3l', 'e0'], 1.0, rtol=0, atol=1e-14)
    assert_allclose(tt['k0sl', 'e0'], 0.1, rtol=0, atol=1e-14)
    assert_allclose(tt['k1sl', 'e0'], 0.2, rtol=0, atol=1e-14)
    assert_allclose(tt['k2sl', 'e0'], 0.3, rtol=0, atol=1e-14)
    assert_allclose(tt['k3sl', 'e0'], 0.4, rtol=0, atol=1e-14)

    assert tt['element_type','e1'] == 'Quadrupole'
    assert tt['isreplica', 'e1'] == False
    assert tt['parent_name', 'e1'] is None
    assert tt['isthick', 'e1'] == True
    assert tt['iscollective', 'e1'] == False
    assert_allclose(tt['s', 'e1'], 1., rtol=0, atol=1e-14)
    assert_allclose(tt['length', 'e1'], 0.5, rtol=0, atol=1e-14)
    assert_allclose(tt['angle_rad', 'e1'], 0.0, rtol=0, atol=1e-14)
    assert_allclose(tt['rot_s_rad', 'e1'], 0.1, rtol=0, atol=1e-14)
    assert_allclose(tt['shift_x', 'e1'], 2e-3, rtol=0, atol=1e-14)
    assert_allclose(tt['shift_y', 'e1'], 1e-3, rtol=0, atol=1e-14)
    assert_allclose(tt['shift_s', 'e1'], -1e-3, rtol=0, atol=1e-14)
    assert_allclose(tt['k0l', 'e1'], 0.0, rtol=0, atol=1e-14)
    assert_allclose(tt['k1l', 'e1'], 0.1 * 0.5, rtol=0, atol=1e-14)
    assert_allclose(tt['k2l', 'e1'], 0.0, rtol=0, atol=1e-14)
    assert_allclose(tt['k3l', 'e1'], 0.0, rtol=0, atol=1e-14)
    assert_allclose(tt['k0sl', 'e1'], 0.0, rtol=0, atol=1e-14)
    assert_allclose(tt['k1sl', 'e1'], 0.2 * 0.5, rtol=0, atol=1e-14)
    assert_allclose(tt['k2sl', 'e1'], 0.0, rtol=0, atol=1e-14)
    assert_allclose(tt['k3sl', 'e1'], 0.0, rtol=0, atol=1e-14)

    assert tt['element_type','e2'] == 'Sextupole'
    assert tt['isreplica', 'e2'] == False
    assert tt['parent_name', 'e2'] is None
    assert tt['isthick', 'e2'] == True
    assert tt['iscollective', 'e2'] == False
    assert_allclose(tt['s', 'e2'], 1.5, rtol=0, atol=1e-14)
    assert_allclose(tt['length', 'e2'], 0.3, rtol=0, atol=1e-14)
    assert_allclose(tt['angle_rad', 'e2'], 0.0, rtol=0, atol=1e-14)
    assert_allclose(tt['rot_s_rad', 'e2'], 0.3, rtol=0, atol=1e-14)
    assert_allclose(tt['shift_x', 'e2'], 3e-3, rtol=0, atol=1e-14)
    assert_allclose(tt['shift_y', 'e2'], 4e-3, rtol=0, atol=1e-14)
    assert_allclose(tt['shift_s', 'e2'], -2e-3, rtol=0, atol=1e-14)
    assert_allclose(tt['k0l', 'e2'], 0.0, rtol=0, atol=1e-14)
    assert_allclose(tt['k1l', 'e2'], 0.0, rtol=0, atol=1e-14)
    assert_allclose(tt['k2l', 'e2'], 0.1 * 0.3, rtol=0, atol=1e-14)
    assert_allclose(tt['k3l', 'e2'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['k0sl', 'e2'], 0.0, rtol=0, atol=1e-14)
    assert_allclose(tt['k1sl', 'e2'], 0.0, rtol=0, atol=1e-14)
    assert_allclose(tt['k2sl', 'e2'], 0.2 * 0.3, rtol=0, atol=1e-14)
    assert_allclose(tt['k3sl', 'e2'], 0, rtol=0, atol=1e-14)

    assert tt['element_type','e3'] == 'Octupole'
    assert tt['isreplica', 'e3'] == False
    assert tt['parent_name', 'e3'] is None
    assert tt['isthick', 'e3'] == True
    assert tt['iscollective', 'e3'] == False
    assert_allclose(tt['s', 'e3'], 1.8, rtol=0, atol=1e-14)
    assert_allclose(tt['length', 'e3'], 0.4, rtol=0, atol=1e-14)
    assert_allclose(tt['angle_rad', 'e3'], 0.0, rtol=0, atol=1e-14)
    assert_allclose(tt['rot_s_rad', 'e3'], 0.4, rtol=0, atol=1e-14)
    assert_allclose(tt['shift_x', 'e3'], 5e-3, rtol=0, atol=1e-14)
    assert_allclose(tt['shift_y', 'e3'], 6e-3, rtol=0, atol=1e-14)
    assert_allclose(tt['shift_s', 'e3'], -1e-3, rtol=0, atol=1e-14)
    assert_allclose(tt['k0l', 'e3'], 0.0, rtol=0, atol=1e-14)
    assert_allclose(tt['k1l', 'e3'], 0.0, rtol=0, atol=1e-14)
    assert_allclose(tt['k2l', 'e3'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['k3l', 'e3'], 0.1 * 0.4, rtol=0, atol=1e-14)
    assert_allclose(tt['k0sl', 'e3'], 0.0, rtol=0, atol=1e-14)
    assert_allclose(tt['k1sl', 'e3'], 0.0, rtol=0, atol=1e-14)
    assert_allclose(tt['k2sl', 'e3'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['k3sl', 'e3'], 0.2 * 0.4, rtol=0, atol=1e-14)

    assert tt['element_type','e4'] == 'Multipole'
    assert tt['isreplica', 'e4'] == False
    assert tt['parent_name', 'e4'] is None
    assert tt['isthick', 'e4'] == False
    assert tt['iscollective', 'e4'] == False
    assert_allclose(tt['s', 'e4'], 2.2, rtol=0, atol=1e-14)
    assert_allclose(tt['length', 'e4'], 0.4, rtol=0, atol=1e-14)
    assert_allclose(tt['angle_rad', 'e4'], 0.1, rtol=0, atol=1e-14)
    assert_allclose(tt['rot_s_rad', 'e4'], 0.7, rtol=0, atol=1e-14)
    assert_allclose(tt['shift_x', 'e4'], 5e-3, rtol=0, atol=1e-14)
    assert_allclose(tt['shift_y', 'e4'], 6e-3, rtol=0, atol=1e-14)
    assert_allclose(tt['shift_s', 'e4'], -1e-3, rtol=0, atol=1e-14)
    assert_allclose(tt['k0l', 'e4'], 0.7, rtol=0, atol=1e-14)
    assert_allclose(tt['k1l', 'e4'], 0.8, rtol=0, atol=1e-14)
    assert_allclose(tt['k2l', 'e4'], 0.9, rtol=0, atol=1e-14)
    assert_allclose(tt['k3l', 'e4'], 1.0, rtol=0, atol=1e-14)
    assert_allclose(tt['k0sl', 'e4'], 0.1, rtol=0, atol=1e-14)
    assert_allclose(tt['k1sl', 'e4'], 0.2, rtol=0, atol=1e-14)
    assert_allclose(tt['k2sl', 'e4'], 0.3, rtol=0, atol=1e-14)
    assert_allclose(tt['k3sl', 'e4'], 0.4, rtol=0, atol=1e-14)

    assert tt['element_type','e5'] == 'Drift'
    assert tt['isreplica', 'e5'] == False
    assert tt['parent_name', 'e5'] is None
    assert tt['isthick', 'e5'] == True
    assert tt['iscollective', 'e5'] == False
    assert_allclose(tt['s', 'e5'], 2.2, rtol=0, atol=1e-14)
    assert_allclose(tt['length', 'e5'], 5.0, rtol=0, atol=1e-14)
    assert_allclose(tt['angle_rad', 'e5'], 0.0, rtol=0, atol=1e-14)
    assert_allclose(tt['rot_s_rad', 'e5'], 0.0, rtol=0, atol=1e-14)
    assert_allclose(tt['shift_x', 'e5'], 0.0, rtol=0, atol=1e-14)
    assert_allclose(tt['shift_y', 'e5'], 0.0, rtol=0, atol=1e-14)
    assert_allclose(tt['shift_s', 'e5'], 0.0, rtol=0, atol=1e-14)
    assert_allclose(tt['k0l', 'e5'], 0.0, rtol=0, atol=1e-14)
    assert_allclose(tt['k1l', 'e5'], 0.0, rtol=0, atol=1e-14)
    assert_allclose(tt['k2l', 'e5'], 0.0, rtol=0, atol=1e-14)
    assert_allclose(tt['k3l', 'e5'], 0.0, rtol=0, atol=1e-14)
    assert_allclose(tt['k0sl', 'e5'], 0.0, rtol=0, atol=1e-14)
    assert_allclose(tt['k1sl', 'e5'], 0.0, rtol=0, atol=1e-14)
    assert_allclose(tt['k2sl', 'e5'], 0.0, rtol=0, atol=1e-14)
    assert_allclose(tt['k3sl', 'e5'], 0.0, rtol=0, atol=1e-14)

    assert tt['element_type','e6'] == 'Bend'
    assert tt['isreplica', 'e6'] == True
    assert tt['parent_name', 'e6'] == 'e0'
    assert tt['isthick', 'e6'] == True
    assert tt['iscollective', 'e6'] == False
    assert_allclose(tt['length', 'e6'], 1, rtol=0, atol=1e-14)
    assert_allclose(tt['angle_rad', 'e6'], 0.3, rtol=0, atol=1e-14)
    assert_allclose(tt['rot_s_rad', 'e6'], 0.2, rtol=0, atol=1e-14)
    assert_allclose(tt['shift_x', 'e6'], 1e-3, rtol=0, atol=1e-14)
    assert_allclose(tt['shift_y', 'e6'], 2e-3, rtol=0, atol=1e-14)
    assert_allclose(tt['shift_s', 'e6'], 2e-3, rtol=0, atol=1e-14)
    assert_allclose(tt['k0l', 'e6'], 0.4 * 1 + 0.7, rtol=0, atol=1e-14)
    assert_allclose(tt['k1l', 'e6'], 0.1 * 1 + 0.8, rtol=0, atol=1e-14)
    assert_allclose(tt['k2l', 'e6'], 0.9, rtol=0, atol=1e-14)
    assert_allclose(tt['k3l', 'e6'], 1.0, rtol=0, atol=1e-14)
    assert_allclose(tt['k0sl', 'e6'], 0.1, rtol=0, atol=1e-14)
    assert_allclose(tt['k1sl', 'e6'], 0.2, rtol=0, atol=1e-14)
    assert_allclose(tt['k2sl', 'e6'], 0.3, rtol=0, atol=1e-14)
    assert_allclose(tt['k3sl', 'e6'], 0.4, rtol=0, atol=1e-14)


@for_all_test_contexts
@pytest.mark.parametrize(
    "check_expr",
    [False, True],
    ids=['no_expr', 'with_expr'],
)
@pytest.mark.parametrize(
    "use_copy",
    [False, True],
    ids=['no_copy', 'with_copy'],
)
def test_attr_thin_slicing(test_context, check_expr, use_copy):

    if check_expr:
        line = _make_line_with_expressions()
    else:
        line = _make_line_no_expressions()

    line.slice_thick_elements(
        slicing_strategies=[xt.Strategy(xt.Uniform(2, mode='thin'))])

    line.build_tracker(_context=test_context)

    if use_copy:
        line = line.copy()
        line.build_tracker(_context=test_context)

    if check_expr:
        _set_vars(line)

    tt = line.get_table(attr=True)

    assert_allclose(tt['s', -1], 8.2, rtol=0, atol=1e-14)

    assert tt['element_type','e0..1'] == 'ThinSliceBend'
    assert tt['isreplica', 'e0..1'] == False
    assert tt['parent_name', 'e0..1'] == 'e0'
    assert tt['isthick', 'e0..1'] == False
    assert tt['iscollective', 'e0..1'] == False
    assert_allclose(tt['s', 'e0..1'], 2./3., rtol=0, atol=1e-14)
    assert_allclose(tt['length', 'e0..1'], 1/2, rtol=0, atol=1e-14)
    assert_allclose(tt['angle_rad', 'e0..1'], 0.3/2, rtol=0, atol=1e-14)
    assert_allclose(tt['rot_s_rad', 'e0..1'], 0.2, rtol=0, atol=1e-14)
    assert_allclose(tt['shift_x', 'e0..1'], 1e-3, rtol=0, atol=1e-14)
    assert_allclose(tt['shift_y', 'e0..1'], 2e-3, rtol=0, atol=1e-14)
    assert_allclose(tt['shift_s', 'e0..1'], 2e-3, rtol=0, atol=1e-14)
    assert_allclose(tt['k0l', 'e0..1'], 0.5 * (0.4 * 1 + 0.7), rtol=0, atol=1e-14)
    assert_allclose(tt['k1l', 'e0..1'], 0.5 * (0.1 * 1 + 0.8), rtol=0, atol=1e-14)
    assert_allclose(tt['k2l', 'e0..1'], 0.5 *  .9, rtol=0, atol=1e-14)
    assert_allclose(tt['k3l', 'e0..1'], 0.5 * 1., rtol=0, atol=1e-14)
    assert_allclose(tt['k0sl', 'e0..1'], 0.5 * 0.1, rtol=0, atol=1e-14)
    assert_allclose(tt['k1sl', 'e0..1'], 0.5 * 0.2, rtol=0, atol=1e-14)
    assert_allclose(tt['k2sl', 'e0..1'], 0.5 *  0.3, rtol=0, atol=1e-14)
    assert_allclose(tt['k3sl', 'e0..1'], 0.5 *  0.4, rtol=0, atol=1e-14)

    assert tt['element_type','e0..exit_map'] == 'ThinSliceBendExit'
    assert tt['isreplica', 'e0..exit_map'] == False
    assert tt['parent_name', 'e0..exit_map'] == 'e0'
    assert tt['isthick', 'e0..exit_map'] == False
    assert tt['iscollective', 'e0..exit_map'] == False
    assert_allclose(tt['s', 'e0..exit_map'], 1., rtol=0, atol=1e-14)
    assert_allclose(tt['length', 'e0..exit_map'], 0., rtol=0, atol=1e-14)
    assert_allclose(tt['angle_rad', 'e0..exit_map'], 0., rtol=0, atol=1e-14)
    assert_allclose(tt['rot_s_rad', 'e0..exit_map'], 0.2, rtol=0, atol=1e-14)
    assert_allclose(tt['shift_x', 'e0..exit_map'], 1e-3, rtol=0, atol=1e-14)
    assert_allclose(tt['shift_y', 'e0..exit_map'], 2e-3, rtol=0, atol=1e-14)
    assert_allclose(tt['shift_s', 'e0..exit_map'], 2e-3, rtol=0, atol=1e-14)
    assert_allclose(tt['k0l', 'e0..exit_map'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['k1l', 'e0..exit_map'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['k2l', 'e0..exit_map'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['k3l', 'e0..exit_map'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['k0sl', 'e0..exit_map'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['k1sl', 'e0..exit_map'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['k2sl', 'e0..exit_map'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['k3sl', 'e0..exit_map'], 0, rtol=0, atol=1e-14)

    assert tt['element_type','e0..entry_map'] == 'ThinSliceBendEntry'
    assert tt['isreplica', 'e0..entry_map'] == False
    assert tt['parent_name', 'e0..entry_map'] == 'e0'
    assert tt['isthick', 'e0..entry_map'] == False
    assert tt['iscollective', 'e0..entry_map'] == False
    assert_allclose(tt['s', 'e0..entry_map'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['length', 'e0..entry_map'], 0., rtol=0, atol=1e-14)
    assert_allclose(tt['angle_rad', 'e0..entry_map'], 0., rtol=0, atol=1e-14)
    assert_allclose(tt['rot_s_rad', 'e0..entry_map'], 0.2, rtol=0, atol=1e-14)
    assert_allclose(tt['shift_x', 'e0..entry_map'], 1e-3, rtol=0, atol=1e-14)
    assert_allclose(tt['shift_y', 'e0..entry_map'], 2e-3, rtol=0, atol=1e-14)
    assert_allclose(tt['shift_s', 'e0..entry_map'], 2e-3, rtol=0, atol=1e-14)
    assert_allclose(tt['k0l', 'e0..entry_map'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['k1l', 'e0..entry_map'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['k2l', 'e0..entry_map'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['k3l', 'e0..entry_map'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['k0sl', 'e0..entry_map'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['k1sl', 'e0..entry_map'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['k2sl', 'e0..entry_map'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['k3sl', 'e0..entry_map'], 0, rtol=0, atol=1e-14)

    assert tt['element_type','drift_e0..1'] == 'DriftSliceBend'
    assert tt['isreplica', 'drift_e0..1'] == False
    assert tt['parent_name', 'drift_e0..1'] == 'e0'
    assert tt['isthick', 'drift_e0..1'] == True
    assert tt['iscollective', 'drift_e0..1'] == False
    assert_allclose(tt['s', 'drift_e0..1'], 1./3, rtol=0, atol=1e-14)
    assert_allclose(tt['length', 'drift_e0..1'], 1/3, rtol=0, atol=1e-14)
    assert_allclose(tt['angle_rad', 'drift_e0..1'], 0., rtol=0, atol=1e-14)
    assert_allclose(tt['rot_s_rad', 'drift_e0..1'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['shift_x', 'drift_e0..1'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['shift_y', 'drift_e0..1'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['shift_s', 'drift_e0..1'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['k0l', 'drift_e0..1'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['k1l', 'drift_e0..1'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['k2l', 'drift_e0..1'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['k3l', 'drift_e0..1'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['k0sl', 'drift_e0..1'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['k1sl', 'drift_e0..1'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['k2sl', 'drift_e0..1'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['k3sl', 'drift_e0..1'], 0, rtol=0, atol=1e-14)

    assert tt['element_type','e1..1'] == 'ThinSliceQuadrupole'
    assert tt['isreplica', 'e1..1'] == False
    assert tt['parent_name', 'e1..1'] == 'e1'
    assert tt['isthick', 'e1..1'] == False
    assert tt['iscollective', 'e1..1'] == False
    assert_allclose(tt['s', 'e1..1'], 1 + .5*2./3., rtol=0, atol=1e-14)
    assert_allclose(tt['length', 'e1..1'], 0.5*1/2, rtol=0, atol=1e-14)
    assert_allclose(tt['angle_rad', 'e1..1'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['rot_s_rad', 'e1..1'], 0.1, rtol=0, atol=1e-14)
    assert_allclose(tt['shift_x', 'e1..1'], 2e-3, rtol=0, atol=1e-14)
    assert_allclose(tt['shift_y', 'e1..1'], 1e-3, rtol=0, atol=1e-14)
    assert_allclose(tt['shift_s', 'e1..1'], -1e-3, rtol=0, atol=1e-14)
    assert_allclose(tt['k0l', 'e1..1'], 0., rtol=0, atol=1e-14)
    assert_allclose(tt['k1l', 'e1..1'], 0.1 * 0.5 / 2, rtol=0, atol=1e-14)
    assert_allclose(tt['k2l', 'e1..1'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['k3l', 'e1..1'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['k0sl', 'e1..1'], 0., rtol=0, atol=1e-14)
    assert_allclose(tt['k1sl', 'e1..1'], 0.5 * 0.2 / 2, rtol=0, atol=1e-14)
    assert_allclose(tt['k2sl', 'e1..1'], 0., rtol=0, atol=1e-14)
    assert_allclose(tt['k3sl', 'e1..1'], 0., rtol=0, atol=1e-14)

    assert tt['element_type','drift_e1..1'] == 'DriftSliceQuadrupole'
    assert tt['isreplica', 'drift_e1..1'] == False
    assert tt['parent_name', 'drift_e1..1'] == 'e1'
    assert tt['isthick', 'drift_e1..1'] == True
    assert tt['iscollective', 'drift_e1..1'] == False
    assert_allclose(tt['s', 'drift_e1..1'], 1 + 0.5 * 1./3, rtol=0, atol=1e-14)
    assert_allclose(tt['length', 'drift_e1..1'], 0.5 * 1/3, rtol=0, atol=1e-14)
    assert_allclose(tt['angle_rad', 'drift_e1..1'], 0., rtol=0, atol=1e-14)
    assert_allclose(tt['rot_s_rad', 'drift_e1..1'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['shift_x', 'drift_e1..1'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['shift_y', 'drift_e1..1'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['shift_s', 'drift_e1..1'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['k0l', 'drift_e1..1'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['k1l', 'drift_e1..1'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['k2l', 'drift_e1..1'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['k3l', 'drift_e1..1'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['k0sl', 'drift_e1..1'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['k1sl', 'drift_e1..1'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['k2sl', 'drift_e1..1'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['k3sl', 'drift_e1..1'], 0, rtol=0, atol=1e-14)

    assert tt['element_type','e1_entry'] == 'Marker'
    assert tt['isreplica', 'e1_entry'] == False
    assert tt['parent_name', 'e1_entry'] is None
    assert tt['isthick', 'e1_entry'] == False
    assert tt['iscollective', 'e1_entry'] == False
    assert_allclose(tt['s', 'e1_entry'], 1, rtol=0, atol=1e-14)
    assert_allclose(tt['length', 'e1_entry'], 0., rtol=0, atol=1e-14)
    assert_allclose(tt['angle_rad', 'e1_entry'], 0., rtol=0, atol=1e-14)
    assert_allclose(tt['rot_s_rad', 'e1_entry'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['shift_x', 'e1_entry'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['shift_y', 'e1_entry'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['shift_s', 'e1_entry'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['k0l', 'e1_entry'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['k1l', 'e1_entry'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['k2l', 'e1_entry'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['k3l', 'e1_entry'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['k0sl', 'e1_entry'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['k1sl', 'e1_entry'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['k2sl', 'e1_entry'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['k3sl', 'e1_entry'], 0, rtol=0, atol=1e-14)

    assert tt['element_type','e2..1'] == 'ThinSliceSextupole'
    assert tt['isreplica', 'e2..1'] == False
    assert tt['parent_name', 'e2..1'] == 'e2'
    assert tt['isthick', 'e2..1'] == False
    assert tt['iscollective', 'e2..1'] == False
    assert_allclose(tt['s', 'e2..1'], 1.5 + .3*2./3., rtol=0, atol=1e-14)
    assert_allclose(tt['length', 'e2..1'], 0.3*1/2, rtol=0, atol=1e-14)
    assert_allclose(tt['angle_rad', 'e2..1'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['rot_s_rad', 'e2..1'], 0.3, rtol=0, atol=1e-14)
    assert_allclose(tt['shift_x', 'e2..1'], 3e-3, rtol=0, atol=1e-14)
    assert_allclose(tt['shift_y', 'e2..1'], 4e-3, rtol=0, atol=1e-14)
    assert_allclose(tt['shift_s', 'e2..1'], -2e-3, rtol=0, atol=1e-14)
    assert_allclose(tt['k0l', 'e2..1'], 0., rtol=0, atol=1e-14)
    assert_allclose(tt['k1l', 'e2..1'], 0., rtol=0, atol=1e-14)
    assert_allclose(tt['k2l', 'e2..1'], 0.1 * 0.3*1/2, rtol=0, atol=1e-14)
    assert_allclose(tt['k3l', 'e2..1'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['k0sl', 'e2..1'], 0., rtol=0, atol=1e-14)
    assert_allclose(tt['k1sl', 'e2..1'], 0., rtol=0, atol=1e-14)
    assert_allclose(tt['k2sl', 'e2..1'], 0.2 * 0.3*1/2, rtol=0, atol=1e-14)
    assert_allclose(tt['k3sl', 'e2..1'], 0., rtol=0, atol=1e-14)

    assert tt['element_type','drift_e2..1'] == 'DriftSliceSextupole'
    assert tt['isreplica', 'drift_e2..1'] == False
    assert tt['parent_name', 'drift_e2..1'] == 'e2'
    assert tt['isthick', 'drift_e2..1'] == True
    assert tt['iscollective', 'drift_e2..1'] == False
    assert_allclose(tt['s', 'drift_e2..1'], 1.5 + .3*1./3., rtol=0, atol=1e-14)
    assert_allclose(tt['length', 'drift_e2..1'], 0.3*1/3, rtol=0, atol=1e-14)
    assert_allclose(tt['angle_rad', 'drift_e2..1'], 0., rtol=0, atol=1e-14)
    assert_allclose(tt['rot_s_rad', 'drift_e2..1'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['shift_x', 'drift_e2..1'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['shift_y', 'drift_e2..1'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['shift_s', 'drift_e2..1'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['k0l', 'drift_e2..1'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['k1l', 'drift_e2..1'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['k2l', 'drift_e2..1'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['k3l', 'drift_e2..1'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['k0sl', 'drift_e2..1'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['k1sl', 'drift_e2..1'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['k2sl', 'drift_e2..1'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['k3sl', 'drift_e2..1'], 0, rtol=0, atol=1e-14)

    assert tt['element_type','e3..1'] == 'ThinSliceOctupole'
    assert tt['isreplica', 'e3..1'] == False
    assert tt['parent_name', 'e3..1'] == 'e3'
    assert tt['isthick', 'e3..1'] == False
    assert tt['iscollective', 'e3..1'] == False
    assert_allclose(tt['s', 'e3..1'], 1.8 + .4*2./3., rtol=0, atol=1e-14)
    assert_allclose(tt['length', 'e3..1'], 0.4*1/2, rtol=0, atol=1e-14)
    assert_allclose(tt['angle_rad', 'e3..1'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['rot_s_rad', 'e3..1'], 0.4, rtol=0, atol=1e-14)
    assert_allclose(tt['shift_x', 'e3..1'], 5e-3, rtol=0, atol=1e-14)
    assert_allclose(tt['shift_y', 'e3..1'], 6e-3, rtol=0, atol=1e-14)
    assert_allclose(tt['shift_s', 'e3..1'], -1e-3, rtol=0, atol=1e-14)
    assert_allclose(tt['k0l', 'e3..1'], 0., rtol=0, atol=1e-14)
    assert_allclose(tt['k1l', 'e3..1'], 0., rtol=0, atol=1e-14)
    assert_allclose(tt['k2l', 'e3..1'], 0., rtol=0, atol=1e-14)
    assert_allclose(tt['k3l', 'e3..1'], 0.1 * 0.4*1/2, rtol=0, atol=1e-14)
    assert_allclose(tt['k0sl', 'e3..1'], 0., rtol=0, atol=1e-14)
    assert_allclose(tt['k1sl', 'e3..1'], 0., rtol=0, atol=1e-14)
    assert_allclose(tt['k2sl', 'e3..1'], 0., rtol=0, atol=1e-14)
    assert_allclose(tt['k3sl', 'e3..1'], 0.2 * 0.4*1/2, rtol=0, atol=1e-14)

    assert tt['element_type','drift_e3..1'] == 'DriftSliceOctupole'
    assert tt['isreplica', 'drift_e3..1'] == False
    assert tt['parent_name', 'drift_e3..1'] == 'e3'
    assert tt['isthick', 'drift_e3..1'] == True
    assert tt['iscollective', 'drift_e3..1'] == False
    assert_allclose(tt['s', 'drift_e3..1'], 1.8 + .4*1./3., rtol=0, atol=1e-14)
    assert_allclose(tt['length', 'drift_e3..1'], 0.4*1/3, rtol=0, atol=1e-14)
    assert_allclose(tt['angle_rad', 'drift_e3..1'], 0., rtol=0, atol=1e-14)
    assert_allclose(tt['rot_s_rad', 'drift_e3..1'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['shift_x', 'drift_e3..1'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['shift_y', 'drift_e3..1'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['shift_s', 'drift_e3..1'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['k0l', 'drift_e3..1'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['k1l', 'drift_e3..1'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['k2l', 'drift_e3..1'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['k3l', 'drift_e3..1'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['k0sl', 'drift_e3..1'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['k1sl', 'drift_e3..1'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['k2sl', 'drift_e3..1'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['k3sl', 'drift_e3..1'], 0, rtol=0, atol=1e-14)

    # Check e5 untouched
    assert tt['element_type','e5'] == 'Drift'
    assert tt['isreplica', 'e5'] == False
    assert tt['parent_name', 'e5'] is None
    assert tt['isthick', 'e5'] == True
    assert tt['iscollective', 'e5'] == False
    assert_allclose(tt['s', 'e5'], 2.2, rtol=0, atol=1e-14)
    assert_allclose(tt['length', 'e5'], 5.0, rtol=0, atol=1e-14)
    assert_allclose(tt['angle_rad', 'e5'], 0.0, rtol=0, atol=1e-14)
    assert_allclose(tt['rot_s_rad', 'e5'], 0.0, rtol=0, atol=1e-14)
    assert_allclose(tt['shift_x', 'e5'], 0.0, rtol=0, atol=1e-14)
    assert_allclose(tt['shift_y', 'e5'], 0.0, rtol=0, atol=1e-14)
    assert_allclose(tt['shift_s', 'e5'], 0.0, rtol=0, atol=1e-14)
    assert_allclose(tt['k0l', 'e5'], 0.0, rtol=0, atol=1e-14)
    assert_allclose(tt['k1l', 'e5'], 0.0, rtol=0, atol=1e-14)
    assert_allclose(tt['k2l', 'e5'], 0.0, rtol=0, atol=1e-14)
    assert_allclose(tt['k3l', 'e5'], 0.0, rtol=0, atol=1e-14)
    assert_allclose(tt['k0sl', 'e5'], 0.0, rtol=0, atol=1e-14)
    assert_allclose(tt['k1sl', 'e5'], 0.0, rtol=0, atol=1e-14)
    assert_allclose(tt['k2sl', 'e5'], 0.0, rtol=0, atol=1e-14)
    assert_allclose(tt['k3sl', 'e5'], 0.0, rtol=0, atol=1e-14)

    assert tt['element_type','e6..1'] == 'ThinSliceBend'
    assert tt['isreplica', 'e6..1'] == False
    assert tt['parent_name', 'e6..1'] == 'e0'
    assert tt['isthick', 'e6..1'] == False
    assert tt['iscollective', 'e6..1'] == False
    assert_allclose(tt['s', 'e6..1'], 7.2 + 2./3., rtol=0, atol=1e-14)
    assert_allclose(tt['length', 'e6..1'], 1/2, rtol=0, atol=1e-14)
    assert_allclose(tt['angle_rad', 'e6..1'], 0.3/2, rtol=0, atol=1e-14)
    assert_allclose(tt['rot_s_rad', 'e6..1'], 0.2, rtol=0, atol=1e-14)
    assert_allclose(tt['shift_x', 'e6..1'], 1e-3, rtol=0, atol=1e-14)
    assert_allclose(tt['shift_y', 'e6..1'], 2e-3, rtol=0, atol=1e-14)
    assert_allclose(tt['shift_s', 'e6..1'], 2e-3, rtol=0, atol=1e-14)
    assert_allclose(tt['k0l', 'e6..1'], 0.5 * (0.4 * 1 + 0.7), rtol=0, atol=1e-14)
    assert_allclose(tt['k1l', 'e6..1'], 0.5 * (0.1 * 1 + 0.8), rtol=0, atol=1e-14)
    assert_allclose(tt['k2l', 'e6..1'], 0.5 *  .9, rtol=0, atol=1e-14)
    assert_allclose(tt['k3l', 'e6..1'], 0.5 * 1., rtol=0, atol=1e-14)
    assert_allclose(tt['k0sl', 'e6..1'], 0.5 * 0.1, rtol=0, atol=1e-14)
    assert_allclose(tt['k1sl', 'e6..1'], 0.5 * 0.2, rtol=0, atol=1e-14)
    assert_allclose(tt['k2sl', 'e6..1'], 0.5 *  0.3, rtol=0, atol=1e-14)
    assert_allclose(tt['k3sl', 'e6..1'], 0.5 *  0.4, rtol=0, atol=1e-14)

    assert tt['element_type','e6..exit_map'] == 'ThinSliceBendExit'
    assert tt['isreplica', 'e6..exit_map'] == False
    assert tt['parent_name', 'e6..exit_map'] == 'e0'
    assert tt['isthick', 'e6..exit_map'] == False
    assert tt['iscollective', 'e6..exit_map'] == False
    assert_allclose(tt['s', 'e6..exit_map'], 7.2 + 1., rtol=0, atol=1e-14)
    assert_allclose(tt['length', 'e6..exit_map'], 0., rtol=0, atol=1e-14)
    assert_allclose(tt['angle_rad', 'e6..exit_map'], 0., rtol=0, atol=1e-14)
    assert_allclose(tt['rot_s_rad', 'e6..exit_map'], 0.2, rtol=0, atol=1e-14)
    assert_allclose(tt['shift_x', 'e6..exit_map'], 1e-3, rtol=0, atol=1e-14)
    assert_allclose(tt['shift_y', 'e6..exit_map'], 2e-3, rtol=0, atol=1e-14)
    assert_allclose(tt['shift_s', 'e6..exit_map'], 2e-3, rtol=0, atol=1e-14)
    assert_allclose(tt['k0l', 'e6..exit_map'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['k1l', 'e6..exit_map'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['k2l', 'e6..exit_map'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['k3l', 'e6..exit_map'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['k0sl', 'e6..exit_map'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['k1sl', 'e6..exit_map'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['k2sl', 'e6..exit_map'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['k3sl', 'e6..exit_map'], 0, rtol=0, atol=1e-14)

    assert tt['element_type','e6..entry_map'] == 'ThinSliceBendEntry'
    assert tt['isreplica', 'e6..entry_map'] == False
    assert tt['parent_name', 'e6..entry_map'] == 'e0'
    assert tt['isthick', 'e6..entry_map'] == False
    assert tt['iscollective', 'e6..entry_map'] == False
    assert_allclose(tt['s', 'e6..entry_map'], 7.2, rtol=0, atol=1e-14)
    assert_allclose(tt['length', 'e6..entry_map'], 0., rtol=0, atol=1e-14)
    assert_allclose(tt['angle_rad', 'e6..entry_map'], 0., rtol=0, atol=1e-14)
    assert_allclose(tt['rot_s_rad', 'e6..entry_map'], 0.2, rtol=0, atol=1e-14)
    assert_allclose(tt['shift_x', 'e6..entry_map'], 1e-3, rtol=0, atol=1e-14)
    assert_allclose(tt['shift_y', 'e6..entry_map'], 2e-3, rtol=0, atol=1e-14)
    assert_allclose(tt['shift_s', 'e6..entry_map'], 2e-3, rtol=0, atol=1e-14)
    assert_allclose(tt['k0l', 'e6..entry_map'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['k1l', 'e6..entry_map'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['k2l', 'e6..entry_map'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['k3l', 'e6..entry_map'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['k0sl', 'e6..entry_map'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['k1sl', 'e6..entry_map'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['k2sl', 'e6..entry_map'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['k3sl', 'e6..entry_map'], 0, rtol=0, atol=1e-14)

    assert tt['element_type','drift_e6..1'] == 'DriftSliceBend'
    assert tt['isreplica', 'drift_e6..1'] == False
    assert tt['parent_name', 'drift_e6..1'] == 'e0'
    assert tt['isthick', 'drift_e6..1'] == True
    assert tt['iscollective', 'drift_e6..1'] == False
    assert_allclose(tt['s', 'drift_e6..1'], 7.2 + 1./3, rtol=0, atol=1e-14)
    assert_allclose(tt['length', 'drift_e6..1'], 1/3, rtol=0, atol=1e-14)
    assert_allclose(tt['angle_rad', 'drift_e6..1'], 0., rtol=0, atol=1e-14)
    assert_allclose(tt['rot_s_rad', 'drift_e6..1'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['shift_x', 'drift_e6..1'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['shift_y', 'drift_e6..1'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['shift_s', 'drift_e6..1'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['k0l', 'drift_e6..1'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['k1l', 'drift_e6..1'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['k2l', 'drift_e6..1'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['k3l', 'drift_e6..1'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['k0sl', 'drift_e6..1'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['k1sl', 'drift_e6..1'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['k2sl', 'drift_e6..1'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['k3sl', 'drift_e6..1'], 0, rtol=0, atol=1e-14)


@for_all_test_contexts
@pytest.mark.parametrize(
    "check_expr",
    [False, True],
    ids=["no_expr", "with_expr"],
)
@pytest.mark.parametrize(
    "use_copy",
    [False, True],
    ids=["no_copy", "with_copy"],
)
def test_attr_thick_slicing(test_context, check_expr, use_copy):

    if check_expr:
        line = _make_line_with_expressions()
    else:
        line = _make_line_no_expressions()

    line.slice_thick_elements(
        slicing_strategies=[xt.Strategy(xt.Uniform(2, mode='thick'))])

    line.build_tracker(_context=test_context)

    if use_copy:
        line = line.copy()
        line.build_tracker(_context=test_context)

    if check_expr:
        _set_vars(line)

    tt = line.get_table(attr=True)

    assert_allclose(tt['s', -1], 8.2, rtol=0, atol=1e-14)

    assert tt['element_type','e0..1'] == 'ThickSliceBend'
    assert tt['isreplica', 'e0..1'] == False
    assert tt['parent_name', 'e0..1'] == 'e0'
    assert tt['isthick', 'e0..1'] == True
    assert tt['iscollective', 'e0..1'] == False
    assert_allclose(tt['s', 'e0..1'], 1./2., rtol=0, atol=1e-14)
    assert_allclose(tt['length', 'e0..1'], 1/2, rtol=0, atol=1e-14)
    assert_allclose(tt['angle_rad', 'e0..1'], 0.3/2, rtol=0, atol=1e-14)
    assert_allclose(tt['rot_s_rad', 'e0..1'], 0.2, rtol=0, atol=1e-14)
    assert_allclose(tt['shift_x', 'e0..1'], 1e-3, rtol=0, atol=1e-14)
    assert_allclose(tt['shift_y', 'e0..1'], 2e-3, rtol=0, atol=1e-14)
    assert_allclose(tt['shift_s', 'e0..1'], 2e-3, rtol=0, atol=1e-14)
    assert_allclose(tt['k0l', 'e0..1'], 0.5 * (0.4 * 1 + 0.7), rtol=0, atol=1e-14)
    assert_allclose(tt['k1l', 'e0..1'], 0.5 * (0.1 * 1 + 0.8), rtol=0, atol=1e-14)
    assert_allclose(tt['k2l', 'e0..1'], 0.5 *  .9, rtol=0, atol=1e-14)
    assert_allclose(tt['k3l', 'e0..1'], 0.5 * 1., rtol=0, atol=1e-14)
    assert_allclose(tt['k0sl', 'e0..1'], 0.5 * 0.1, rtol=0, atol=1e-14)
    assert_allclose(tt['k1sl', 'e0..1'], 0.5 * 0.2, rtol=0, atol=1e-14)
    assert_allclose(tt['k2sl', 'e0..1'], 0.5 *  0.3, rtol=0, atol=1e-14)
    assert_allclose(tt['k3sl', 'e0..1'], 0.5 *  0.4, rtol=0, atol=1e-14)

    assert tt['element_type','e0..exit_map'] == 'ThinSliceBendExit'
    assert tt['isreplica', 'e0..exit_map'] == False
    assert tt['parent_name', 'e0..exit_map'] == 'e0'
    assert tt['isthick', 'e0..exit_map'] == False
    assert tt['iscollective', 'e0..exit_map'] == False
    assert_allclose(tt['s', 'e0..exit_map'], 1., rtol=0, atol=1e-14)
    assert_allclose(tt['length', 'e0..exit_map'], 0., rtol=0, atol=1e-14)
    assert_allclose(tt['angle_rad', 'e0..exit_map'], 0., rtol=0, atol=1e-14)
    assert_allclose(tt['rot_s_rad', 'e0..exit_map'], 0.2, rtol=0, atol=1e-14)
    assert_allclose(tt['shift_x', 'e0..exit_map'], 1e-3, rtol=0, atol=1e-14)
    assert_allclose(tt['shift_y', 'e0..exit_map'], 2e-3, rtol=0, atol=1e-14)
    assert_allclose(tt['shift_s', 'e0..exit_map'], 2e-3, rtol=0, atol=1e-14)
    assert_allclose(tt['k0l', 'e0..exit_map'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['k1l', 'e0..exit_map'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['k2l', 'e0..exit_map'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['k3l', 'e0..exit_map'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['k0sl', 'e0..exit_map'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['k1sl', 'e0..exit_map'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['k2sl', 'e0..exit_map'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['k3sl', 'e0..exit_map'], 0, rtol=0, atol=1e-14)

    assert tt['element_type','e0..entry_map'] == 'ThinSliceBendEntry'
    assert tt['isreplica', 'e0..entry_map'] == False
    assert tt['parent_name', 'e0..entry_map'] == 'e0'
    assert tt['isthick', 'e0..entry_map'] == False
    assert tt['iscollective', 'e0..entry_map'] == False
    assert_allclose(tt['s', 'e0..entry_map'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['length', 'e0..entry_map'], 0., rtol=0, atol=1e-14)
    assert_allclose(tt['angle_rad', 'e0..entry_map'], 0., rtol=0, atol=1e-14)
    assert_allclose(tt['rot_s_rad', 'e0..entry_map'], 0.2, rtol=0, atol=1e-14)
    assert_allclose(tt['shift_x', 'e0..entry_map'], 1e-3, rtol=0, atol=1e-14)
    assert_allclose(tt['shift_y', 'e0..entry_map'], 2e-3, rtol=0, atol=1e-14)
    assert_allclose(tt['shift_s', 'e0..entry_map'], 2e-3, rtol=0, atol=1e-14)
    assert_allclose(tt['k0l', 'e0..entry_map'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['k1l', 'e0..entry_map'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['k2l', 'e0..entry_map'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['k3l', 'e0..entry_map'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['k0sl', 'e0..entry_map'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['k1sl', 'e0..entry_map'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['k2sl', 'e0..entry_map'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['k3sl', 'e0..entry_map'], 0, rtol=0, atol=1e-14)

    assert tt['element_type','e1..1'] == 'ThickSliceQuadrupole'
    assert tt['isreplica', 'e1..1'] == False
    assert tt['parent_name', 'e1..1'] == 'e1'
    assert tt['isthick', 'e1..1'] == True
    assert tt['iscollective', 'e1..1'] == False
    assert_allclose(tt['s', 'e1..1'], 1 + .5 / 2, rtol=0, atol=1e-14)
    assert_allclose(tt['length', 'e1..1'], 0.5*1/2, rtol=0, atol=1e-14)
    assert_allclose(tt['angle_rad', 'e1..1'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['rot_s_rad', 'e1..1'], 0.1, rtol=0, atol=1e-14)
    assert_allclose(tt['shift_x', 'e1..1'], 2e-3, rtol=0, atol=1e-14)
    assert_allclose(tt['shift_y', 'e1..1'], 1e-3, rtol=0, atol=1e-14)
    assert_allclose(tt['shift_s', 'e1..1'], -1e-3, rtol=0, atol=1e-14)
    assert_allclose(tt['k0l', 'e1..1'], 0., rtol=0, atol=1e-14)
    assert_allclose(tt['k1l', 'e1..1'], 0.1 * 0.5 / 2, rtol=0, atol=1e-14)
    assert_allclose(tt['k2l', 'e1..1'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['k3l', 'e1..1'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['k0sl', 'e1..1'], 0., rtol=0, atol=1e-14)
    assert_allclose(tt['k1sl', 'e1..1'], 0.5 * 0.2 / 2, rtol=0, atol=1e-14)
    assert_allclose(tt['k2sl', 'e1..1'], 0., rtol=0, atol=1e-14)
    assert_allclose(tt['k3sl', 'e1..1'], 0., rtol=0, atol=1e-14)

    assert tt['element_type','e1_entry'] == 'Marker'
    assert tt['isreplica', 'e1_entry'] == False
    assert tt['parent_name', 'e1_entry'] is None
    assert tt['isthick', 'e1_entry'] == False
    assert tt['iscollective', 'e1_entry'] == False
    assert_allclose(tt['s', 'e1_entry'], 1, rtol=0, atol=1e-14)
    assert_allclose(tt['length', 'e1_entry'], 0., rtol=0, atol=1e-14)
    assert_allclose(tt['angle_rad', 'e1_entry'], 0., rtol=0, atol=1e-14)
    assert_allclose(tt['rot_s_rad', 'e1_entry'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['shift_x', 'e1_entry'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['shift_y', 'e1_entry'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['shift_s', 'e1_entry'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['k0l', 'e1_entry'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['k1l', 'e1_entry'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['k2l', 'e1_entry'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['k3l', 'e1_entry'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['k0sl', 'e1_entry'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['k1sl', 'e1_entry'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['k2sl', 'e1_entry'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['k3sl', 'e1_entry'], 0, rtol=0, atol=1e-14)

    assert tt['element_type','e2..1'] == 'ThickSliceSextupole'
    assert tt['isreplica', 'e2..1'] == False
    assert tt['parent_name', 'e2..1'] == 'e2'
    assert tt['isthick', 'e2..1'] == True
    assert tt['iscollective', 'e2..1'] == False
    assert_allclose(tt['s', 'e2..1'], 1.5 + .3 / 2., rtol=0, atol=1e-14)
    assert_allclose(tt['length', 'e2..1'], 0.3*1/2, rtol=0, atol=1e-14)
    assert_allclose(tt['angle_rad', 'e2..1'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['rot_s_rad', 'e2..1'], 0.3, rtol=0, atol=1e-14)
    assert_allclose(tt['shift_x', 'e2..1'], 3e-3, rtol=0, atol=1e-14)
    assert_allclose(tt['shift_y', 'e2..1'], 4e-3, rtol=0, atol=1e-14)
    assert_allclose(tt['shift_s', 'e2..1'], -2e-3, rtol=0, atol=1e-14)
    assert_allclose(tt['k0l', 'e2..1'], 0., rtol=0, atol=1e-14)
    assert_allclose(tt['k1l', 'e2..1'], 0., rtol=0, atol=1e-14)
    assert_allclose(tt['k2l', 'e2..1'], 0.1 * 0.3*1/2, rtol=0, atol=1e-14)
    assert_allclose(tt['k3l', 'e2..1'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['k0sl', 'e2..1'], 0., rtol=0, atol=1e-14)
    assert_allclose(tt['k1sl', 'e2..1'], 0., rtol=0, atol=1e-14)
    assert_allclose(tt['k2sl', 'e2..1'], 0.2 * 0.3*1/2, rtol=0, atol=1e-14)
    assert_allclose(tt['k3sl', 'e2..1'], 0., rtol=0, atol=1e-14)

    assert tt['element_type','e3..1'] == 'ThickSliceOctupole'
    assert tt['isreplica', 'e3..1'] == False
    assert tt['parent_name', 'e3..1'] == 'e3'
    assert tt['isthick', 'e3..1'] == True
    assert tt['iscollective', 'e3..1'] == False
    assert_allclose(tt['s', 'e3..1'], 1.8 + .4 /2., rtol=0, atol=1e-14)
    assert_allclose(tt['length', 'e3..1'], 0.4*1/2, rtol=0, atol=1e-14)
    assert_allclose(tt['angle_rad', 'e3..1'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['rot_s_rad', 'e3..1'], 0.4, rtol=0, atol=1e-14)
    assert_allclose(tt['shift_x', 'e3..1'], 5e-3, rtol=0, atol=1e-14)
    assert_allclose(tt['shift_y', 'e3..1'], 6e-3, rtol=0, atol=1e-14)
    assert_allclose(tt['shift_s', 'e3..1'], -1e-3, rtol=0, atol=1e-14)
    assert_allclose(tt['k0l', 'e3..1'], 0., rtol=0, atol=1e-14)
    assert_allclose(tt['k1l', 'e3..1'], 0., rtol=0, atol=1e-14)
    assert_allclose(tt['k2l', 'e3..1'], 0., rtol=0, atol=1e-14)
    assert_allclose(tt['k3l', 'e3..1'], 0.1 * 0.4*1/2, rtol=0, atol=1e-14)
    assert_allclose(tt['k0sl', 'e3..1'], 0., rtol=0, atol=1e-14)
    assert_allclose(tt['k1sl', 'e3..1'], 0., rtol=0, atol=1e-14)
    assert_allclose(tt['k2sl', 'e3..1'], 0., rtol=0, atol=1e-14)
    assert_allclose(tt['k3sl', 'e3..1'], 0.2 * 0.4*1/2, rtol=0, atol=1e-14)

    # Check e5 untouched
    assert tt['element_type','e5'] == 'Drift'
    assert tt['isreplica', 'e5'] == False
    assert tt['parent_name', 'e5'] is None
    assert tt['isthick', 'e5'] == True
    assert tt['iscollective', 'e5'] == False
    assert_allclose(tt['s', 'e5'], 2.2, rtol=0, atol=1e-14)
    assert_allclose(tt['length', 'e5'], 5.0, rtol=0, atol=1e-14)
    assert_allclose(tt['angle_rad', 'e5'], 0.0, rtol=0, atol=1e-14)
    assert_allclose(tt['rot_s_rad', 'e5'], 0.0, rtol=0, atol=1e-14)
    assert_allclose(tt['shift_x', 'e5'], 0.0, rtol=0, atol=1e-14)
    assert_allclose(tt['shift_y', 'e5'], 0.0, rtol=0, atol=1e-14)
    assert_allclose(tt['shift_s', 'e5'], 0.0, rtol=0, atol=1e-14)
    assert_allclose(tt['k0l', 'e5'], 0.0, rtol=0, atol=1e-14)
    assert_allclose(tt['k1l', 'e5'], 0.0, rtol=0, atol=1e-14)
    assert_allclose(tt['k2l', 'e5'], 0.0, rtol=0, atol=1e-14)
    assert_allclose(tt['k3l', 'e5'], 0.0, rtol=0, atol=1e-14)
    assert_allclose(tt['k0sl', 'e5'], 0.0, rtol=0, atol=1e-14)
    assert_allclose(tt['k1sl', 'e5'], 0.0, rtol=0, atol=1e-14)
    assert_allclose(tt['k2sl', 'e5'], 0.0, rtol=0, atol=1e-14)
    assert_allclose(tt['k3sl', 'e5'], 0.0, rtol=0, atol=1e-14)

    assert tt['element_type','e6..1'] == 'ThickSliceBend'
    assert tt['isreplica', 'e6..1'] == False
    assert tt['parent_name', 'e6..1'] == 'e0'
    assert tt['isthick', 'e6..1'] == True
    assert tt['iscollective', 'e6..1'] == False
    assert_allclose(tt['s', 'e6..1'], 7.2 + 1./2., rtol=0, atol=1e-14)
    assert_allclose(tt['length', 'e6..1'], 1/2, rtol=0, atol=1e-14)
    assert_allclose(tt['angle_rad', 'e6..1'], 0.3/2, rtol=0, atol=1e-14)
    assert_allclose(tt['rot_s_rad', 'e6..1'], 0.2, rtol=0, atol=1e-14)
    assert_allclose(tt['shift_x', 'e6..1'], 1e-3, rtol=0, atol=1e-14)
    assert_allclose(tt['shift_y', 'e6..1'], 2e-3, rtol=0, atol=1e-14)
    assert_allclose(tt['shift_s', 'e6..1'], 2e-3, rtol=0, atol=1e-14)
    assert_allclose(tt['k0l', 'e6..1'], 0.5 * (0.4 * 1 + 0.7), rtol=0, atol=1e-14)
    assert_allclose(tt['k1l', 'e6..1'], 0.5 * (0.1 * 1 + 0.8), rtol=0, atol=1e-14)
    assert_allclose(tt['k2l', 'e6..1'], 0.5 *  .9, rtol=0, atol=1e-14)
    assert_allclose(tt['k3l', 'e6..1'], 0.5 * 1., rtol=0, atol=1e-14)
    assert_allclose(tt['k0sl', 'e6..1'], 0.5 * 0.1, rtol=0, atol=1e-14)
    assert_allclose(tt['k1sl', 'e6..1'], 0.5 * 0.2, rtol=0, atol=1e-14)
    assert_allclose(tt['k2sl', 'e6..1'], 0.5 *  0.3, rtol=0, atol=1e-14)
    assert_allclose(tt['k3sl', 'e6..1'], 0.5 *  0.4, rtol=0, atol=1e-14)

    assert tt['element_type','e6..exit_map'] == 'ThinSliceBendExit'
    assert tt['isreplica', 'e6..exit_map'] == False
    assert tt['parent_name', 'e6..exit_map'] == 'e0'
    assert tt['isthick', 'e6..exit_map'] == False
    assert tt['iscollective', 'e6..exit_map'] == False
    assert_allclose(tt['s', 'e6..exit_map'], 7.2 + 1., rtol=0, atol=1e-14)
    assert_allclose(tt['length', 'e6..exit_map'], 0., rtol=0, atol=1e-14)
    assert_allclose(tt['angle_rad', 'e6..exit_map'], 0., rtol=0, atol=1e-14)
    assert_allclose(tt['rot_s_rad', 'e6..exit_map'], 0.2, rtol=0, atol=1e-14)
    assert_allclose(tt['shift_x', 'e6..exit_map'], 1e-3, rtol=0, atol=1e-14)
    assert_allclose(tt['shift_y', 'e6..exit_map'], 2e-3, rtol=0, atol=1e-14)
    assert_allclose(tt['shift_s', 'e6..exit_map'], 2e-3, rtol=0, atol=1e-14)
    assert_allclose(tt['k0l', 'e6..exit_map'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['k1l', 'e6..exit_map'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['k2l', 'e6..exit_map'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['k3l', 'e6..exit_map'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['k0sl', 'e6..exit_map'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['k1sl', 'e6..exit_map'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['k2sl', 'e6..exit_map'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['k3sl', 'e6..exit_map'], 0, rtol=0, atol=1e-14)

    assert tt['element_type','e6..entry_map'] == 'ThinSliceBendEntry'
    assert tt['isreplica', 'e6..entry_map'] == False
    assert tt['parent_name', 'e6..entry_map'] == 'e0'
    assert tt['isthick', 'e6..entry_map'] == False
    assert tt['iscollective', 'e6..entry_map'] == False
    assert_allclose(tt['s', 'e6..entry_map'], 7.2 + 0, rtol=0, atol=1e-14)
    assert_allclose(tt['length', 'e6..entry_map'], 0., rtol=0, atol=1e-14)
    assert_allclose(tt['angle_rad', 'e6..entry_map'], 0., rtol=0, atol=1e-14)
    assert_allclose(tt['rot_s_rad', 'e6..entry_map'], 0.2, rtol=0, atol=1e-14)
    assert_allclose(tt['shift_x', 'e6..entry_map'], 1e-3, rtol=0, atol=1e-14)
    assert_allclose(tt['shift_y', 'e6..entry_map'], 2e-3, rtol=0, atol=1e-14)
    assert_allclose(tt['shift_s', 'e6..entry_map'], 2e-3, rtol=0, atol=1e-14)
    assert_allclose(tt['k0l', 'e6..entry_map'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['k1l', 'e6..entry_map'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['k2l', 'e6..entry_map'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['k3l', 'e6..entry_map'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['k0sl', 'e6..entry_map'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['k1sl', 'e6..entry_map'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['k2sl', 'e6..entry_map'], 0, rtol=0, atol=1e-14)
    assert_allclose(tt['k3sl', 'e6..entry_map'], 0, rtol=0, atol=1e-14)



import numpy as np
import xtrack as xt
import xobjects as xo
from cpymad.madx import Madx

def test_rbend_rbarc():
    mad = Madx()
    mad.input('''
        ang = 0.6;
        lb = 0.5;
        rb: rbend, l=lb, angle=ang;

        beam, particle=proton, energy=1.0;
        seq: sequence, l=2;
        rb, at=1;
        endsequence;
        use, sequence=seq;
    ''')
    tw = xt.Table(mad.twiss(betx=1.0, bety=1.0))

    ds_madx = tw['s', 'rb:1'] - tw['s', 'rb:1<<1']

    line = xt.Line.from_madx_sequence(mad.sequence.seq)

    xo.assert_allclose(ds_madx, line['rb'].length, atol=0, rtol=1e-12)

    env = xt.Environment()
    env['lb'] = 0.5
    env['ang'] = 0.6
    env.new('rb_rbarc', xt.RBend)
    env.set('rb_rbarc', length_straight='lb', angle='ang', k0_from_h=True)

    xo.assert_allclose(env['rb_rbarc'].length, ds_madx, atol=0, rtol=1e-12)
    xo.assert_allclose(env['rb_rbarc'].h * env['rb_rbarc'].length, 0.6, atol=0, rtol=1e-12)
    xo.assert_allclose(env['rb_rbarc'].edge_entry_angle, 0, atol=0, rtol=1e-12)
    xo.assert_allclose(env['rb_rbarc'].edge_exit_angle, 0, atol=0, rtol=1e-12)
    xo.assert_allclose(env['rb_rbarc'].k0, env['rb_rbarc'].h, atol=0, rtol=1e-12)

    assert env['rb_rbarc'].get_expr('length_straight') == "vars['lb']"
    assert env['rb_rbarc'].get_expr('angle') == "vars['ang']"
    assert env['rb_rbarc'].get_expr('edge_entry_angle') == None
    assert env['rb_rbarc'].get_expr('edge_exit_angle') == None

    env.new('rb_norbarc', xt.Bend)
    env.set('rb_norbarc', length='lb', angle='ang', k0_from_h=True)

    xo.assert_allclose(env['rb_norbarc'].length, 0.5, atol=0, rtol=1e-12)
    xo.assert_allclose(env['rb_norbarc'].h * env['rb_norbarc'].length, 0.6, atol=0, rtol=1e-12)
    xo.assert_allclose(env['rb_norbarc'].k0, env['rb_norbarc'].h, atol=0, rtol=1e-12)
    xo.assert_allclose(env['rb_norbarc'].edge_entry_angle, 0, atol=0, rtol=1e-12)
    xo.assert_allclose(env['rb_norbarc'].edge_exit_angle, 0, atol=0, rtol=1e-12)

    assert env['rb_norbarc'].get_expr('length') == "vars['lb']"
    assert env['rb_norbarc'].get_expr('angle') == "vars['ang']"
    assert env['rb_norbarc'].get_expr('edge_entry_angle') == None
    assert env['rb_norbarc'].get_expr('edge_exit_angle') == None

    # Check an sbend
    env.new('sb', xt.Bend)
    env.set('sb', length='lb', angle='ang', k0_from_h=True)

    xo.assert_allclose(env['sb'].length, 0.5, atol=0, rtol=1e-12)
    xo.assert_allclose(env['sb'].h * env['sb'].length, 0.6, atol=0, rtol=1e-12)
    xo.assert_allclose(env['sb'].edge_entry_angle, 0, atol=1e-12, rtol=0)
    xo.assert_allclose(env['sb'].edge_exit_angle, 0, atol=1e-12, rtol=0)

    assert env['sb'].get_expr('length') == "vars['lb']"
    assert env['sb'].get_expr('angle') == "vars['ang']"
    assert env['sb'].get_expr('k0') == env['sb'].get_expr('h')
    assert env['sb'].get_expr('edge_entry_angle') == None
    assert env['sb'].get_expr('edge_exit_angle') == None
